import sys
import warnings

import grpc
import numpy as np
import pandas as pd
from radiens.api.api_utils.util import (parse_stim_param_resp,
                                        stim_params_dict_to_req)
from radiens.exceptions.grpc_error import handle_grpc_error
from radiens.grpc_radiens import (allegoserver_pb2, allegoserver_pb2_grpc,
                                  biointerface_pb2, common_pb2, datasource_pb2,
                                  spikesorter_pb2)
from radiens.lib.allego_lib import AllegoState
from radiens.lib.channel_metadata import ChannelMetadata
from radiens.lib.sig_metrics import SignalMetrics, SignalMetricsStatus
from radiens.lib.spike_sorter import Dashboard, SorterState
from radiens.lib.spikes import SpikesMetadata, SpikesSet
from radiens.utils.config import new_server_channel
from radiens.utils.constants import (DEFAULT_STIM_PARAMS, HEADSTAGE_ALIAS,
                                     PRIMARY_CACHE_STREAM_GROUP_ID,
                                     PROBE_ALIAS, SPIKE_SORTER_ID)
from radiens.utils.enums import (ClientType, DioMode, Port, RecordMode,
                                 StreamMode, SystemMode, WorkspaceApp)

# ====== life cycle ======


def restart(addr, mode: SystemMode):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.Restart(allegoserver_pb2.RestartRequest(
                mode=mode.enum))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)

# ====== workspace ======


def workspace_save(addr, is_force_overwrite=True, tags='', notes=''):
    annotate = common_pb2.AnnotateBundle(tags=tags, notes=notes)
    req = common_pb2.WorkspaceControlRequest(cmd=common_pb2.WSPACE_Save,
                                             annotation=annotate, isForceOverwrite=is_force_overwrite)
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.WorkspaceControl(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)


def workspace_save_as(addr, wspace_id: str, is_force_overwrite=True, tags='', notes=''):
    annotate = common_pb2.AnnotateBundle(tags=tags, notes=notes)
    req = common_pb2.WorkspaceControlRequest(cmd=common_pb2.WSPACE_Save,
                                             workspaceID=wspace_id,
                                             annotation=annotate, isForceOverwrite=is_force_overwrite)
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.WorkspaceControl(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)


def workspace_switch(addr, wspace_id: str):
    req = common_pb2.WorkspaceControlRequest(cmd=common_pb2.WSPACE_Switch,
                                             workspaceID=wspace_id)
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.WorkspaceControl(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)


def workspace_delete(addr, wspace_id: str):
    req = common_pb2.WorkspaceControlRequest(cmd=common_pb2.WSPACE_Delete,
                                             workspaceID=wspace_id)
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.WorkspaceControl(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)


def workspace_current(addr):
    req = common_pb2.GetWorkspaceRequest(
        cmd=common_pb2.GET_WSPACE_Current, appMask=common_pb2.Allego)
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            resp = stub.GetWorkspace(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        d = {'ID': [], 'app': [], 'last_used': [], 'last_modified': [],
             'is_modified': [], 'notes': [], 'tags': []}
        for _, v in resp.workspaceDesc.items():
            d['ID'].append(v.iD)
            d['app'].append(WorkspaceApp.parse(v.app).name)
            d['last_used'].append(v.timestampLastUsed)
            d['last_modified'].append(v.timestampModified)
            d['is_modified'].append(v.isModified)
            d['tags'].append(v.annotation.tags)
            d['notes'].append(v.annotation.notes)
        return pd.DataFrame(d)


def workspace_list(addr):
    req = common_pb2.GetWorkspaceRequest(
        cmd=common_pb2.GET_WSPACE_List, appMask=common_pb2.Allego)
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            resp = stub.GetWorkspace(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        d = {'ID': [], 'app': [], 'last_used': [], 'last_modified': [],
             'is_modified': [], 'notes': [], 'tags': []}
        for _, v in resp.workspaceDesc.items():
            d['ID'].append(v.iD)
            d['app'].append(WorkspaceApp.parse(v.app).name)
            d['last_used'].append(v.timestampLastUsed)
            d['last_modified'].append(v.timestampModified)
            d['is_modified'].append(v.isModified)
            d['tags'].append(v.annotation.tags)
            d['notes'].append(v.annotation.notes)
        return pd.DataFrame(d)

# ====== getters ======


def get_stream_loop_dur_ms(addr) -> float:
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            res = stub.GetConfig(common_pb2.StandardRequest())
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        return res.streamLoopDurMs


def get_sensors(addr):
    req = common_pb2.SignalGroupIDRequest()
    req.streamGroupId = PRIMARY_CACHE_STREAM_GROUP_ID
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            res = stub.ListSensorSpecs(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        hstages = {}
        for _, v in enumerate(res.headstages):
            hstages[v.name.strip('hstg__')] = v.channelCount
        probes = {}
        for _, v in enumerate(res.probes):
            probes[v.name.strip('probe__')] = v.channelCount
        return {'headstages': hstages, 'probes': probes}


def get_cfg_status(addr):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            res = stub.GetConfigAndStatus(common_pb2.StandardRequest())
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        return AllegoState(res)


def get_signal_group(addr, stream_group_id=None):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            req = common_pb2.SignalGroupIDRequest()
            req.streamGroupId = stream_group_id

            raw = stub.GetSignalGroup(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        return ChannelMetadata(raw)


def get_signals(addr, stream_group_id: str, samp_freq: float):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Pcache1Stub(chan)
        req = common_pb2.GetSignalsRequest(streamGroupId=stream_group_id)
        req.params.wdwSec = 1  # no resampling
        req.params.chanHeightPx = 0  # ain scaling hack
        req.params.ampFsrUv = sys.float_info.max  # no clipping
        req.params.plotWidthPoints = samp_freq  # no resampling
        req.params.gpioOnTop = False
        req.params.ainFsrV = 1
        req.params.componentID = 'allego_python_client'
        req.params.isHeatmap = True  # no amp scaling or offset
        req.params.clipToRange = False
        req.params.resampleType = common_pb2.ResampleRoutine.NAIVE

        try:
            raw = stub.GetSignals(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)

        sigarray = np.frombuffer(raw.data, dtype=np.float32)
        sigarray = np.reshape(sigarray, (raw.shape[0], raw.shape[1]))
        time_range = [raw.timeRange[0], raw.timeRange[1]]

        return sigarray, time_range


def get_digital_out_states(addr):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            res = stub.GetDIOReg(common_pb2.StandardRequest())
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        else:
            return {
                'digital_outs_mode': DioMode.parse(res.mode).name,
                'states': [{'chan_idx': i.ntvChanIdx, 'state': i.manualState} for i in res.doutChanRegisters]
            }


def get_dac_register(addr):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            res = stub.GetDACReg(common_pb2.StandardRequest())
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        ch = []
        for chan in res.analogOutChannels:
            ch.append({'amp_ntv_chan_idx': chan.PriNtvChanIdx,
                       'aout_ntv_chan_idx': chan.AnalogOutNtvChanIdx,
                       'stream': chan.Stream,
                       'stream_offset_idx': chan.StreamOffsetIdx})
        return {
            'dac': ch,
            'gain': res.gain,
            'high_pass': {'enable': res.highpassReg.enable,
                          'cutoff_freq': res.highpassReg.cutoffFreq}}


def get_stim_params(addr):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            resp = stub.GetStimParams(common_pb2.StandardRequest())
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
    return parse_stim_param_resp(resp)

# ======= setters ========


def set_dac_high_pass(addr, enable: bool, cutoff_freq: float):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.SetDACHighPass(allegoserver_pb2.DACHighPassRegister(enable=enable,
                                                                     cutoffFreq=cutoff_freq))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        return get_dac_register(addr)


def set_digital_out_manual(addr, dout1_state: bool, dout2_state: bool):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)

        req1 = allegoserver_pb2.DIOModeManualRequest()
        req1.chanIdx = 0
        req1.state = dout1_state

        req2 = allegoserver_pb2.DIOModeManualRequest()
        req2.chanIdx = 1
        req2.state = dout2_state

        try:
            stub.SetDIOManual(req1)
            stub.SetDIOManual(req2)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)


def set_fs(addr, fs: int):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.SetConfigCore(
                allegoserver_pb2.SetConfigCoreRequest(sampFreq=float(fs)))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)


def set_sensor(addr, port: Port, hstage: str, probe: str):
    status = get_cfg_status(addr)
    if port not in status.port_num_channels:
        raise ValueError(
            'port {} is not valid or is connected'.format(port.name))

    sensors = get_sensors(addr)
    if hstage in HEADSTAGE_ALIAS.keys():
        pb_hstage = 'hstg__{}'.format(HEADSTAGE_ALIAS[hstage])
    elif hstage in sensors['headstages'].keys():
        pb_hstage = 'hstg__{}'.format(hstage)
    else:
        raise ValueError(
            'headstage {} is not an alias nor an available headstage ID'.format(hstage))

    if probe in PROBE_ALIAS.keys():
        pb_probe = 'probe__{}'.format(PROBE_ALIAS[probe])
    elif probe in sensors['probes'].keys():
        pb_probe = 'probe__{}'.format(probe)
    else:
        raise ValueError(
            'probe {} is not an alias nor an available probe ID'.format(hstage))

    hstg_num_chan = sensors['headstages'][pb_hstage.strip('hstg__')]
    probe_num_chan = sensors['probes'][pb_probe.strip('probe__')]
    if status.port_num_channels[port] != hstg_num_chan:
        warnings.warn('port {} number of channels={} does not match headstage number of channels={}'.format(
                      port.name, status.port_num_channels[port], hstg_num_chan))
    if hstg_num_chan != probe_num_chan:
        warnings.warn('headstage {} number of channels={} does not match probe number of channels={}'.format(
                      sensors['headstages'][pb_hstage.strip('hstg__')], hstg_num_chan, probe_num_chan))

    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.SetSensor(common_pb2.SetSensorRequest(streamGroupId=PRIMARY_CACHE_STREAM_GROUP_ID,
                                                       port=port.enum, headstageId=pb_hstage, probeId=pb_probe))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)


def set_manual_stim_trigger(addr, trigger: int):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.ManualStimTriggerToggle(allegoserver_pb2.ManualStimTriggerToggleRequest(
                trigger=trigger))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)


def set_stim_step(addr, stim_step_enum: int):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.SetStimStep(allegoserver_pb2.StimStepMessage(
                stimStep=stim_step_enum))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)


def set_stim_params(addr, stim_params_req: dict):
    with new_server_channel(addr) as chan:
        if 'stim_sys_chan_idx' not in stim_params_req:
            raise KeyError(
                "Required key stim_sys_chan_idx not found in request.")
        stim_sys_chan_idx = stim_params_req['stim_sys_chan_idx']

        curr_params = None
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            curr_params_raw = stub.GetStimParams(common_pb2.StandardRequest())
            curr_params = parse_stim_param_resp(curr_params_raw)
        except grpc.RpcError as ex:
            warnings.warn(
                "can't retrieve any current stim parameters. Continuing with assuming defaults for parameters unspecified in request")

        curr_chan_params = None
        if curr_params is None or curr_params == {}:
            curr_chan_params = DEFAULT_STIM_PARAMS
        else:
            curr_chan_params = curr_params[stim_sys_chan_idx]

        for param, val in curr_chan_params.items():
            if param not in stim_params_req:
                stim_params_req[param] = val

        req = stim_params_dict_to_req(stim_params_req)
        try:
            stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
            stub.SetStimParams(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)


def set_stream_state(addr, state: StreamMode):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.SetStreamState(
                allegoserver_pb2.SetStreamRequest(mode=state.pb_value))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)


def set_record_state(addr, state: RecordMode):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.SetRecordState(
                allegoserver_pb2.SetRecordRequest(mode=state.pb_value))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)


def set_recording_config(addr, filename: str, filepath: str, index: int, time_stamp: bool):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.SetConfigRecording(allegoserver_pb2.ConfigRecording(baseFileName=filename,
                                                                     baseFilePath=filepath,
                                                                     dataSourceIdx=int(
                                                                         index),
                                                                     timeStamp=time_stamp))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)


def set_time_to_cache_head(addr, stream_group_id=None):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Pcache1Stub(chan)
        req = common_pb2.SignalGroupIDRequest()
        req.streamGroupId = stream_group_id
        try:
            stub.SetTimeRangeToHead(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)

# ==========================
# ====== KPIs ==============
# ==========================


def set_kpi_params(addr, dsource_id,  req):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Kpi1Stub(chan)
        try:
            stub.SetKpiParam(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)


def get_kpi_status(addr):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Kpi1Stub(chan)
        try:
            resp = stub.GetKpiStatus(common_pb2.GetKpiStatusRequest(
                streamGroupId=PRIMARY_CACHE_STREAM_GROUP_ID))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        return SignalMetricsStatus(resp)


def get_kpi_metrics(addr, req):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Kpi1Stub(chan)
        try:
            resp = stub.KpiGetMetrics(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        return SignalMetrics(PRIMARY_CACHE_STREAM_GROUP_ID, resp)


def get_kpi_params(addr):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Kpi1Stub(chan)
        try:
            resp = stub.GetKpiParam(datasource_pb2.DataSourceRequest(
                dsourceID=[PRIMARY_CACHE_STREAM_GROUP_ID]))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        p = {'ntv_idx': [], 'chan_enabled': [], 'thr_activated': [], 'thr': [], 'thr_sd': [], 'thr_wdw': [],
             'shadow': [], 'weak_thr_activated': [],  'weak_thr': [], 'weak_thr_sd': [], 'thr_wdw_pts': [], 'shadow_pts': []}
        for v in resp.rec:
            p['ntv_idx'].append(v.ntvChanIdx)
            p['chan_enabled'].append(v.isEnabled)
            p['thr_activated'].append(v.isSetThr)
            p['thr'].append(np.array(v.thr))
            p['thr_sd'].append(np.array(v.thrSd))
            p['thr_wdw'].append(np.around(np.array(v.thrWdw) * 1000.0, 5))
            p['shadow'].append(np.around(np.array(v.shadow) * 1000.0, 5))
            p['weak_thr_activated'].append(np.array(v.isSetWeakThr))
            p['weak_thr'].append(np.array(v.weakThr))
            p['weak_thr_sd'].append(np.array(v.weakThrSd))
            p['thr_wdw_pts'].append(v.thrWdwPts)
            p['shadow_pts'].append(v.shadowPts)
        return pd.DataFrame(p)

# ==========================
# ====== spike sorter ======
# ==========================


def sorter_cmd(addr, cmd):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Neurons1Stub(chan)
        try:
            stub.SpikeSorterCommand(spikesorter_pb2.SpikeSorterCommandRequest(
                cmd=cmd, subCmd=spikesorter_pb2.SORTER_SUBCMD_NULL, spikeSorterID=''))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)


def sorter_set_params(addr, msg):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Neurons1Stub(chan)
        try:
            stub.SpikeSorterSetParams(msg)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)


def sorter_get_params(addr):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Neurons1Stub(chan)
        try:
            resp = stub.SpikeSorterGetParam(
                spikesorter_pb2.SpikeSorterStandardRequest(spikeSorterID=''))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        p = {'ntv_idx': [], 'chan_enabled': [], 'thr_activated': [], 'thr': [], 'thr_sd': [], 'thr_wdw': [],
             'shadow': [], 'weak_thr': [], 'thr_wdw_pts': [], 'shadow_pts': []}
        for v in resp.rec:
            p['ntv_idx'].append(v.ntvChanIdx)
            p['chan_enabled'].append(v.isEnabled)
            p['thr_activated'].append(v.isSetThr)
            p['thr'].append(np.array(v.thr))
            p['thr_sd'].append(np.array(v.thrSd))
            p['thr_wdw'].append(np.around(np.array(v.thrWdw) * 1000.0, 5))
            p['shadow'].append(np.around(np.array(v.shadow) * 1000.0, 5))
            p['weak_thr'].append(np.array(v.weakThr))
            p['thr_wdw_pts'].append(v.thrWdwPts)
            p['shadow_pts'].append(v.shadowPts)
        return pd.DataFrame(p)


def sorter_get_state(addr):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Neurons1Stub(chan)
        try:
            resp = stub.SpikeSorterGetState(
                spikesorter_pb2.SpikeSorterStandardRequest(spikeSorterID=''))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        return SorterState(resp)


def sorter_get_raster_data(addr, msg):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Neurons1Stub(chan)
        try:
            resp = stub.SpikeSorterGetRasterData(msg)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)

        sigarray = np.frombuffer(resp.GPIOData, dtype=np.float32)
        sigarray = np.reshape(sigarray, (resp.GPIOShape[0], resp.GPIOShape[1]))
        return {'time_range': resp.timeRange, 'timestamp_range': resp.timeStampRange,
                'spikes': {'ntv_chan_idx': resp.spikeTimestampsByChannel.ntvChanIdx,
                           'timestamps': resp.spikeTimestampsByChannel.spikeTimestamps,
                           'labels': resp.spikeTimestampsByChannellabels},
                'gpio_data': sigarray}


def sorter_get_dashboard(addr):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Neurons1Stub(chan)
        try:
            resp = stub.SpikeSorterGetDashboard(
                spikesorter_pb2.SpikeSorterStandardRequest(spikeSorterID=''))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
    return Dashboard(resp)


def get_spikes(addr, req):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Neurons1Stub(chan)
        try:
            resp = stub.BiointerfaceGetSpikesDense(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        return SpikesSet(resp)


def get_neurons(addr, req):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Neurons1Stub(chan)
        try:
            resp = stub.BiointerfaceGetNeurons(
                biointerface_pb2.BiointerfaceGetNeuronsRequest(req))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        return resp


def get_spikes_spec(addr):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Neurons1Stub(chan)
        try:
            resp = stub.SpikesGetSpec(
                spikesorter_pb2.SpikeSorterStandardRequest(spikeSorterID=SPIKE_SORTER_ID))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.ALLEGO)
        return SpikesMetadata(resp)



"""
radiens CLI
"""
from datetime import datetime
from pathlib import Path
from pprint import pprint

import click
import pandas as pd
from radiens.utils.config import ERROR_COLOR_FG
from radiens.utils.util import time_now


@ click.group(short_help='spike sorter commands')
@ click.pass_context
def sorter(ctx):
    '''
    Allego real-time spike sorter commands. 

    This command group has sub-commands to control the spike sorter and its data. 
    '''


@ sorter.command('sort', short_help='turn spike sorter on or off')
@ click.pass_context
@ click.argument('arg', type=click.STRING, nargs=-1)
@ click.option('-v', '--verbose',  type=click.Choice(['brief', 'all'], case_sensitive=False), default='brief', help='detail of status', show_default=True)
def sort(ctx, arg, verbose):
    '''
    Controls the spike sorter state. 

    arg: [[on | off]], no argument will print status
    '''
    if len(arg) == 0:
        sorter_status(ctx, verbose)
        return
    if arg[0] not in ['on', 'off']:
        click.echo(click.style(
            '`arg` must be `on` or `off`', fg=ERROR_COLOR_FG))
        return
    try:
        ctx.obj['allego_client'].spike_sorter().set_sorting(arg[0])
    except Exception as ex:
        click.echo('Error: {}'.format(ex))
        return
    click.echo(
        'ALLEGO real-time spike sorter set {} at {}'.format(arg[0], time_now()))


@ sorter.command('initialize', short_help='initializes spike sorter')
@ click.pass_context
def init(ctx):
    '''
    Stops, clears, and initializes the spike sorter to its default parameters. 
    '''
    try:
        ctx.obj['allego_client'].spike_sorter().initialize()
    except Exception as ex:
        click.echo('Error: {}'.format(ex))
        return
    click.echo(
        'ALLEGO real-time spike sorter initialized at {}'.format(time_now()))


@ sorter.command('rebase', short_help='rebases spike sorter')
@ click.pass_context
def rebase(ctx):
    '''
   Rebases the spike sorter, which clears all spike data and spike templates. 
    '''
    try:
        ctx.obj['allego_client'].spike_sorter().rebase()
    except Exception as ex:
        click.echo('Error: {}'.format(ex))
        return
    click.echo('ALLEGO real-time spike sorter rebased at {}'.format(time_now()))


@ sorter.command('clear-spikes', short_help='clears spikes from spike sorter')
@ click.pass_context
def clear_spikes(ctx):
    '''
   Clears all spike data from the spike sorter, but does not clear it's sorting settings or spike templates.
    '''
    try:
        ctx.obj['allego_client'].spike_sorter().clear_spikes()
    except Exception as ex:
        click.echo('Error: {}'.format(ex))
        return
    click.echo(
        'ALLEGO real-time spike sorter spikes cleared at {}'.format(time_now()))


@ sorter.command('status', short_help='prints spike sorter status')
@ click.pass_context
@ click.option('-v', '--verbose',  type=click.Choice(['brief', 'all'], case_sensitive=False), default='brief', help='detail of status', show_default=True)
def sorter_status(ctx, verbose):
    '''
    Prints the spike sorter status. 
    '''
    try:
        resp = ctx.obj['allego_client'].spike_sorter().get_state()
    except Exception as ex:
        click.echo('Error: {}'.format(ex))
        return
    pprint(resp)


@ sorter.command('params', short_help='prints spike sorter channel parameters')
@ click.pass_context
@ click.option('-v', '--verbose',  type=click.Choice(['brief', 'all'], case_sensitive=False), default='brief', help='detail of parameters', show_default=True)
def sorter_params(ctx, verbose):
    '''
    Prints the spike sorter channel parameters, including threshold levels and threshold states for each channel and also the spike detection window and shadow window.  
    '''
    try:
        resp = ctx.obj['allego_client'].spike_sorter().get_params()
    except Exception as ex:
        click.echo('Error: {}'.format(ex))
        return
    pprint(resp)


@ sorter.command('dashboard', short_help='prints spike sorter dashboard')
@ click.pass_context
@ click.option('-v', '--verbose',  type=click.Choice(['brief', 'all'], case_sensitive=False), default='brief', help='detail of dashboard', show_default=True)
def sorter_dash(ctx, verbose):
    '''
    Prints the spike sorter dashboard. 
    '''
    try:
        resp = ctx.obj['allego_client'].spike_sorter().get_dashboard()
    except Exception as ex:
        click.echo('Error: {}'.format(ex))
        return
    click.echo('\nALLEGO spike sorter dashboard')
    if verbose in ['brief']:
        print(resp.general)
        print('\n')
        return
    if verbose in ['all']:
        click.echo('GENERAL ')
        print(resp.general)
        click.echo('\nSUMMARY STATISTICS \n')
        print(resp.summary_stats)
        click.echo('\nSTATE \n')
        print(resp.state)
        return


@ sorter.group('set', short_help='sets the mutable spike sorter settings')
@ click.pass_context
def sorter_set(ctx):
    '''
    Set Allego spike sorter settings

    This command group has sub-commands to set various types of spike sorter settings and parameters. 
    '''


@ sorter_set.command('threshold', short_help='sets threshold level(s) and activation state')
@ click.pass_context
@ click.option('-nl', '--neg_thr_level',  type=click.FLOAT, default=None, help='negative threshold level', show_default=True)
@ click.option('-pl', '--pos_thr_level',  type=click.FLOAT, default=None, help='positive threshold level', show_default=True)
@ click.option('-s', '--scale',  type=click.Choice(['uV', 'sd'], case_sensitive=False), default='uV', help='threshold scale in uV or SD', show_default=True)
@ click.option('-na', '--neg_thr_act',  type=click.BOOL, default=None, help='negative threshold activation state', show_default=True)
@ click.option('-pa', '--pos_thr_act',  type=click.BOOL, default=None, help='positive threshold activation state', show_default=True)
@ click.option('-i', '--idxs',  type=click.INT, default=None, help='requested channels (-1-> all channels)', show_default=True)
@ click.option('-w', '--weak_thr',  type=click.BOOL, default=False, help='use True to set the weak threshold', show_default=True)
def sorter_set_threshold(ctx, neg_thr_level, pos_thr_level, scale, neg_thr_act, pos_thr_act, idxs, weak_thr):
    '''
    Sets threshold levels and activation states
    '''
    idxs = None if idxs == -1 else idxs
    try:
        level_msg = ctx.obj['allego_client'].spike_sorter().set_threshold_level(
            neg_thr=neg_thr_level, pos_thr=pos_thr_level, scale=scale, ntv_idxs=idxs, weak_thr=weak_thr, channel_meta=None)
        activate_msg = ctx.obj['allego_client'].spike_sorter().set_threshold(
            neg_thr=neg_thr_act, pos_thr=pos_thr_act, ntv_idxs=idxs, weak_thr=weak_thr, channel_meta=None)
    except Exception as ex:
        click.echo('Error: {}'.format(ex))
        return
    click.echo('set thresholds : {}'.format(level_msg))
    click.echo('                 {}'.format(activate_msg))


@ sorter_set.command('window', short_help='sets spike detection window')
@ click.pass_context
@ click.option('-pre', '--pre_thr_ms',  type=click.FLOAT, default=None, help='pre-threshold time [ms]', show_default=True)
@ click.option('-pl', '--post_thr_ms',  type=click.FLOAT, default=None, help='post-threshold time [ms]', show_default=True)
def sorter_set_window(ctx, pre_thr_ms, post_thr_ms):
    '''
    Sets the spike detection window
    '''
    try:
        msg = ctx.obj['allego_client'].spike_sorter().set_spike_window(
            pre_thr_ms=pre_thr_ms, post_thr_ms=post_thr_ms)
    except Exception as ex:
        click.echo('Error: {}'.format(ex))
        return
    click.echo('set window  : {}'.format(msg))


@ sorter_set.command('shadow', short_help='sets shadow window')
@ click.pass_context
@ click.option('-s', '--shadow_ms',  type=click.FLOAT, default=None, help='shadow  time [ms]', show_default=True)
def sorter_set_shadow(ctx, shadow_ms):
    '''
    Sets the shadow window
    '''
    try:
        msg = ctx.obj['allego_client'].spike_sorter().set_spike_shadow(
            shadow_ms=shadow_ms)
    except Exception as ex:
        click.echo('Error: {}'.format(ex))
        return
    click.echo('set shadow  : {}'.format(msg))



"""
radiens CLI
"""
from datetime import datetime
from pathlib import Path
from pprint import pprint

import click
import pandas as pd
from radiens.utils.constants import PRIMARY_CACHE_STREAM_GROUP_ID
from radiens.utils.enums import ClientType
from radiens.utils.util import make_time_range, time_now


@ click.group(short_help='signals commands')
@ click.pass_context
def signals(ctx):
    '''
    Signals commands. 

    This command group has sub-commands for interacting with signals. 
    '''


@ signals.command('snapshot', short_help='get and plot a signals snapshot')
@ click.pass_context
@ click.option('-idx', '--index',  type=click.INT, default=None, help='dataset index from table (Allego n/a)', show_default=True)
@ click.option('-id', '--dset_id',  type=click.STRING, default=None, help='dataset ID from table (Allego n/a)', show_default=True)
@ click.option('-p', '--path',  type=click.STRING, default=None, help='path and base name of Radiens file set', show_default=True)
@ click.option('-h', '--hub',  type=click.STRING, default='default', help='Radiens hub (Allego n/a)', show_default=True)
def snapshot(ctx, index, dset_id, path, hub):
    '''
    Gets a signals snapshot and optionally plot it or save to file. 

    '''
    cmd_app = ClientType.parse(ctx.obj['cmd_app'])
    try:
        if cmd_app.is_videre():
            dsource = ctx.obj['videre_client'].get_data_file_metadata(
                dataset_idx=index, dataset_id=dset_id, path=path, fail_hard=True)
            ctx.obj['videre_client'].signal_metrics().clear(
                dataset_metadata=dsource)
            msg_id = dsource.id
        else:
            click.Echo('not implemented yet for Allego')
            msg_id = PRIMARY_CACHE_STREAM_GROUP_ID
    except Exception as ex:
        click.echo('Error: {}'.format(ex))
        return
    click.echo('{} ({}) : cleared signal metrics at {}'.format(
        cmd_app.name, msg_id, time_now()))


@ signals.command('psd', short_help='calculate power spectral density')
@ click.pass_context
@ click.option('-ts', '--tstart',  type=click.FLOAT, default=None, help='time start (sec)', show_default=True)
@ click.option('-dur', '--dur',  type=click.FLOAT, default=None, help='time end (sec)', show_default=True)
@ click.option('-idx', '--index',  type=click.INT, default=None, help='dataset index from table (Allego n/a)', show_default=True)
@ click.option('-id', '--dset_id',  type=click.STRING, default=None, help='dataset ID from table (Allego n/a)', show_default=True)
@ click.option('-p', '--path',  type=click.STRING, default=None, help='path and base name of Radiens file set', show_default=True)
@ click.option('-f', '--file',  type=click.STRING, default=None, help='save to file', show_default=True)
@ click.option('-h', '--hub',  type=click.STRING, default='default', help='Radiens hub (Allego n/a)', show_default=True)
def psd(ctx, tstart, dur, index, dset_id, path, file, hub):
    '''
    Calculates the power spectral density of the requested signals and optionally plots the result and/or saves it to file. 
    '''
    cmd_app = ClientType.parse(ctx.obj['cmd_app'])
    try:
        if cmd_app.is_videre():
            dsource = ctx.obj['videre_client'].get_data_file_metadata(
                dataset_idx=index, dataset_id=dset_id, path=path, fail_hard=True)
            tstart = tstart if tstart is not None else dsource.time_range.sec[0]
            dur = dur if dur is not None else 0.250
            ctx.obj['videre_client'].signals().get_psd(
                time_range=make_time_range(
                    time_range=[tstart, tstart+dur], fs=dsource.time_range.fs),
                file=str(Path(file).expanduser().absolute()) if file else None,
                data=False,
                dataset_metadata=dsource)
            msg_id = dsource.id
        else:
            click.Echo('not implemented yet for Allego')
            msg_id = PRIMARY_CACHE_STREAM_GROUP_ID
    except Exception as ex:
        click.echo('Error: {}'.format(ex))
        return
    click.echo('{} ({}) : calculated PSD heatmap at {}'.format(
        cmd_app.name, msg_id, time_now()))
    if file:
        click.echo('          saved to {}'.format(file))


# @ signals.command('dashboard', short_help='launch signals dashboard')
# @ click.pass_context
# @ click.option('-idx', '--index',  type=click.INT, default=None, help='dataset index from table', show_default=True)
# @ click.option('-id', '--dset_id',  type=click.STRING, default=None, help='dataset ID from table', show_default=True)
# @ click.option('-p', '--path',  type=click.STRING, default=None, help='path and base name of Radiens file set', show_default=True)
# @ click.option('-d', '--dash',  type=click.Choice(['metrics', 'psd', 'snapshot'], case_sensitive=False), default='metrics', help='dashboard name', show_default=True)
# @ click.option('-s', '--style',  type=click.Choice(['dark', '538', 'fast', 'seaborn'], case_sensitive=False), default='dark', help='dashboard style', show_default=True)
# @ click.option('-did', '--dash_id',  type=click.STRING, default=None, help='radiens-py dash ID', show_default=True)
# @ click.option('-h', '--hub',  type=click.STRING, default='default', help='Radiens hub', show_default=True)
# def dashboard(ctx, index, dset_id, path, dash, style, dash_id, hub):
#     '''
#     Launch a dashboard for a data source linked to the Radiens hub.
#     '''
#     try:
#         dash_enum = DASHBOARD(['metrics', 'psd', 'snapshot'].index(dash))
#         style_enum = DASHBOARD_STYLE(['dark', '538', 'fast', 'seaborn'].index(style))
#         dsource = ctx.obj['videre_client'].get_data_file_metadata(dataset_idx=index, dataset_id=dset_id, path=path, fail_hard=True)
#         ctx.obj['videre_client'].signals().launch_dashboard(dsource, dash_enum, dash_style=style_enum, dash_id=dash_id)
#     except Exception as ex:
#         click.echo('Error: {}'.format(ex))
#         return
#     click.echo('Launching {} dashboard in {} style...'.format(dash_enum.name, style_enum.name))



import warnings

import grpc
import numpy as np
import pandas as pd
from radiens.api.api_utils.protocols import ProtocolAPI, make_protocol_from_pb2
from radiens.exceptions.grpc_error import handle_grpc_error
from radiens.grpc_radiens import (allegoserver_pb2, allegoserver_pb2_grpc,
                                  common_pb2, datasource_pb2,
                                  radiensserver_pb2, radiensserver_pb2_grpc,
                                  spikesorter_pb2)
from radiens.lib.fsys import FileSysResponse
from radiens.utils.config import new_server_channel
from radiens.utils.enums import ClientType

# ====== life cycle ======

# file system


def dsrc_list_dir(addr, dirL: list, sort_by: str):
    sort_enum = ['date', 'name', 'size', 'type']
    if sort_by not in sort_enum:
        raise ValueError('{} must be in {}'.format(sort_by, sort_enum))

    if len(dirL) == 1:
        req = common_pb2.ListDataSourcesRequest(
            directory=dirL[0], sortBy=sort_enum.index(sort_by))
    else:
        req = common_pb2.ListDataSourcesRequest(
            directory='na', dirList=dirL, sortBy=sort_enum.index(sort_by))
    with new_server_channel(addr) as chan:
        stub = radiensserver_pb2_grpc.RadiensCoreStub(chan)
        try:
            resp = stub.ListDirectory(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.CURATE)
        return FileSysResponse(resp)


def dsrc_copy(addr, req: datasource_pb2.CopyRemoveDataSourceFileRequest):
    with new_server_channel(addr) as chan:
        stub = radiensserver_pb2_grpc.RadiensCoreStub(chan)
        try:
            resp = stub.CopyDataSourceFile(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.CURATE)
        return FileSysResponse(resp)


def dsrc_remove(addr, req: datasource_pb2.CopyRemoveDataSourceFileRequest):
    with new_server_channel(addr) as chan:
        stub = radiensserver_pb2_grpc.RadiensCoreStub(chan)
        try:
            resp = stub.RemoveDataSourceFile(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.CURATE)
        return FileSysResponse(resp)


def dsrc_move(addr, req: datasource_pb2.MoveDataSourceFileRequest):
    with new_server_channel(addr) as chan:
        stub = radiensserver_pb2_grpc.RadiensCoreStub(chan)
        try:
            resp = stub.MoveDataSourceFile(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.CURATE)
        return FileSysResponse(resp)


def set_protocol(addr, protocol: ProtocolAPI):
    with new_server_channel(addr) as chan:
        stub = radiensserver_pb2_grpc.RadiensCoreStub(chan)
        try:
            resp = stub.SetProtocol(protocol.as_protobuf())
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.CURATE)
        return make_protocol_from_pb2(resp)


def get_protocol(addr, protocol_id: str):
    with new_server_channel(addr) as chan:
        stub = radiensserver_pb2_grpc.RadiensCoreStub(chan)
        try:
            resp = stub.GetProtocol(
                common_pb2.ProtocolRequest(protocolID=protocol_id))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.CURATE)
        return make_protocol_from_pb2(resp)


def rename_protocol(addr, protocol_id: str, new_protocol_id: str):
    with new_server_channel(addr) as chan:
        stub = radiensserver_pb2_grpc.RadiensCoreStub(chan)
        try:
            resp = stub.RenameProtocol(common_pb2.RenameProtocolRequest(
                protocolID=protocol_id, newProtocolID=new_protocol_id))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.CURATE)
        return make_protocol_from_pb2(resp)


def get_all_protocols(addr):
    with new_server_channel(addr) as chan:
        stub = radiensserver_pb2_grpc.RadiensCoreStub(chan)
        try:
            resp = stub.GetAllProtocols(common_pb2.StandardRequest())
        except grpc.RpcError as ex:
            handle_grpc_error(ex, ClientType.CURATE)
        protocols = []
        for p in resp.protocols:
            protocols.append(make_protocol_from_pb2(p))
        return protocols


def apply_protocol(addr, protocol_id: str):
    chan = new_server_channel(addr)
    stub = radiensserver_pb2_grpc.RadiensCoreStub(chan)
    try:
        stream = stub.ApplyProtocol(
            common_pb2.ProtocolRequest(protocolID=protocol_id))
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.CURATE)
    return stream



from typing import TYPE_CHECKING, Union

import numpy as np
import pandas as pd
from radiens.api import api_allego, api_videre
from radiens.grpc_radiens import common_pb2
from radiens.lib.channel_metadata import ChannelMetadata
from radiens.lib.dataset_metadata import DatasetMetadata
from radiens.lib.sig_metrics import SignalMetrics, SignalMetricsStatus
from radiens.utils.constants import (DEFAULT_HUB_ID, KPI_ADDR, NEURONS1_ADDR,
                                     PRIMARY_CACHE_STREAM_GROUP_ID)
from radiens.utils.enums import (MetricID, MetricMode, MetricName,
                                 RadiensService, SignalType)

# the following lines are to avoid circular imports and are only used for typing hints
# (TYPE_CHECKING always evaluates to false at runtime)
if TYPE_CHECKING:
    from radiens.curate_client import CurateClient
    from radiens.videre_client import VidereClient


class MetricsClient:
    """
    Metrics client object for CurateClient, VidereClient
    """

    def __init__(self, parent_client):
        """ """
        self.__parent: Union[CurateClient, VidereClient] = parent_client

    def clear(self, dataset_metadata=None, hub_name=DEFAULT_HUB_ID) -> None:
        """
        Clears KPI data.  This is only available for :py:class:`.CurateClient` and :py:class:`.VidereClient`

        Parameters:
            dataset_metadata (~radiens.lib.dataset_metadata.DatasetMetadata): dataset metadata object
        Returns:
            None

        Example:
            >>> client.signal_metrics().clear()
            None

        See Also:
            :py:meth:`~radiens.VidereClient.get_data_file_metadata`
            :py:meth:`~radiens.VidereClient.link_data_file`

        """
        if self._is_allego():
            raise RuntimeError("metrics clear is not available for allego")
        else:
            if not isinstance(dataset_metadata, DatasetMetadata):
                raise ValueError(
                    'videre: dataset_metadata must be a DatasetMetadata')
            api_videre.set_kpi_clear(self.__parent._server_address(
                hub_name, RadiensService.CORE), dataset_metadata.attributes['dsource_id'])

    def _is_allego(self) -> bool:
        return self.__parent.type.is_allego()

    def _is_curate(self) -> bool:
        return self.__parent.type.is_curate()

    def _is_videre(self) -> bool:
        return self.__parent.type.is_videre()

    def calculate(self,
                  dataset_metadata=None,
                  hub_name=DEFAULT_HUB_ID) -> None:
        """
        Calculates and saves signal metrics for a linked dataset.
        If ``cmd=standard``, then signal metrics are calculated using the set of standard parameters:
          thresholds:  2.0 * Vex_rms  (negative and positive)

        If ``cmd=standard``, then signal metrics are calculated using the existing signal metrics parameters (thresholds, windows, etc.)

        This is only available for CurateClient and VidereClient

        Parameters:
            dataset_metadata (~radiens.lib.dataset_metadata.DatasetMetadata): dataset metadata object         

        Returns:
            None

        Example:
            >>> client.signal_metrics().calculate()
            None

        See Also:
            :py:meth:`~radiens.VidereClient.get_data_file_metadata`
            :py:meth:`~radiens.VidereClient.link_data_file`

        """
        if self._is_allego():
            raise RuntimeError("metrics calculate is not available for allego")
        if not isinstance(dataset_metadata, DatasetMetadata):
            raise ValueError(
                "videre: dataset_metadata must be a DatasetMetadata")
        if not cmd in ["standard", "custom"]:
            raise ValueError("cmd must be `standard` or `custom`")
        cmd = (
            common_pb2.KPI_SCANNER_STANDARD_SPRINT
            if cmd in ["standard"]
            else common_pb2.KPI_SCANNER_BESPOKE_SPRINT
        )
        req = common_pb2.KpiStandardRequest(
            dsourceID=[dataset_metadata.attributes["dsource_id"]], cmd=cmd
        )
        api_videre.set_kpi_calculate(
            self.__parent._server_address(hub_name, RadiensService.CORE),
            req,
        )

    def set_event_threshold_level(
        self,
        neg_thr=None,
        pos_thr=None,
        scale="uV",
        ntv_idxs=None,
        weak_thr=False,
        dataset_metadata=None,
        hub_name=DEFAULT_HUB_ID,
        channel_meta=None,
    ) -> str:
        """
        Sets the signal metrics event (spike) detection threshold.  This function does not effect the threshold state.

        Parameters:
            neg_thr (str): ``on`` | ``off``
            pos_thr (str): ``on`` | ``off``
            scale (str):  ``uV`` | ``sd``
            ntv_idxs (list): channel native indices (default None=all channels)
            weak_thr (bool): use True to set the weak threshold (default=False)
            dataset_metadata (~radiens.lib.dataset_metadata.DatasetMetadata): dataset metadata object
            channel_metadata (~radiens.lib.channel_metadata.ChannelMetadata): Allego connected channels

        Returns:
            msg (str): summary description of threshold levels

        Example:
            >>> client.signal_metrics().set_threshold_level(neg_thr=85)
            '[32 channels] detect: neg thr=85.00 uV, pos thr=n/a uV'

        See Also:
             :py:meth:`set_threshold`
             :py:meth:`get_channel_metadata`

        """
        if self._is_videre() and not isinstance(
            dataset_metadata, DatasetMetadata
        ):
            raise ValueError("videre: dataset_metadata must be provided")
        if neg_thr in [None] and pos_thr in [None]:
            return 'no new threshold levels were requested so none were changed'
        if channel_metadata in [None]:
            channel_metadata = self.__parent.get_channel_metadata(
            ) if self._is_allego() else dataset_metadata.channel_metadata
        if not isinstance(channel_metadata, ChannelMetadata):
            raise ValueError(
                'channel metadata must be of type ChannelMetadata')
        if isinstance(ntv_idxs, (list, np.ndarray)):
            pass
        elif isinstance(ntv_idxs, (int, float)):
            ntv_idxs = [ntv_idxs]
        elif ntv_idxs in [None]:
            ntv_idxs = channel_metadata.index(SignalType.AMP).ntv
        else:
            raise ValueError("ntv_idxs must be scalar or list or None")
        reg_float = [0, 0]
        reg_bool = [False, False]
        if neg_thr not in [None]:
            reg_float[0] = neg_thr
            reg_bool[0] = True
        if pos_thr not in [None]:
            reg_float[1] = pos_thr
            reg_bool[1] = True
        if scale in ["sd"] and weak_thr in [False]:
            arg = common_pb2.SORTER_THR_LEVEL_SD
        elif scale in ["uV"] and weak_thr in [False]:
            arg = common_pb2.SORTER_THR_LEVEL
        elif scale in ["sd"] and weak_thr in [True]:
            arg = common_pb2.SORTER_WEAK_THR_LEVEL_SD
        elif scale in ["uV"] and weak_thr in [True]:
            arg = common_pb2.SORTER_WEAK_THR_LEVEL
        else:
            raise ValueError(
                'scale must be `uV` | `sd` and/or weak_thr must be True | False')
        req = [common_pb2.SpikeSorterSetParamRequest(spikeSorterID='not_used',
                                                     cmd=arg,
                                                     regFloat64=reg_float,
                                                     regBool=reg_bool,
                                                     ntvChanIdx=ntv_idxs)]

        if self._is_allego():
            req.spikeSorterID = PRIMARY_CACHE_STREAM_GROUP_ID
            api_allego.sorter_set_params(
                NEURONS1_ADDR, common_pb2.SpikeSorterSetParamsRequest(params=req))
        else:
            if not isinstance(dataset_metadata, DatasetMetadata):
                raise ValueError(
                    'videre: dataset_metadata must be a DatasetMetadata')
            req[0].spikeSorterID = dataset_metadata.attributes['dsource_id']
            api_videre.set_kpi_params(self.__parent._server_address(hub_name, RadiensService.CORE),
                                      common_pb2.SpikeSorterSetParamsRequest(params=req))

        return '[{} channels] {}: neg thr={} {}, pos_thr={} {}' .format(len(ntv_idxs),
                                                                        'detect' if arg in [common_pb2.SORTER_THR_LEVEL,
                                                                                            common_pb2.SORTER_THR_LEVEL_SD] else 'weak',
                                                                        '{:.2f}'.format(reg_float[0]) if reg_bool[0] in [
            True] else 'n/a',
            scale,
            '{:.2f}'.format(reg_float[1]) if reg_bool[1] in [
            True] else 'n/a',
            scale)

    def set_event_threshold(self, neg_thr=None, pos_thr=None, ntv_idxs=None, weak_thr=False, dataset_metadata=None, hub_name=DEFAULT_HUB_ID, channel_metadata=None) -> str:
        """
        Sets signal metrics event (spike) detection threshold 'on' or 'off'.

        Parameters:
            neg_thr (str): ``on`` | ``off``
            pos_thr (str): ``on`` | ``off``
            ntv_idxs (list): channel native indices (default None=all channels)
            weak_thr (bool): use True to set the weak threshold (default=False)
            dataset_metadata (~radiens.lib.dataset_metadata.DatasetMetadata): dataset metadata object
            channel_metadata (~radiens.lib.channel_metadata.ChannelMetadata): Allego connected channels

        Returns:
            msg (str): summary description of threshold levels

        Example:
            >>> client.signal_metrics().set_event_threshold(neg_thr='on')
            '[32 channels] detect: neg thr='on', pos thr=n/a'

        See Also:
             :py:meth:`set_threshold_level`
             :py:meth:`get_channel_metadata`

        """
        if not self._is_videre() and not isinstance(
            dataset_metadata, DatasetMetadata
        ):
            raise ValueError("videre: dataset_metadata must be provided")
        if neg_thr in [None] and pos_thr in [None]:
            return 'no new threshold states were requested so none were changed'
        if channel_metadata in [None]:
            channel_metadata = self.__parent.get_channel_metadata(
            ) if self._is_allego() else dataset_metadata.channel_metadata
        if not isinstance(channel_metadata, ChannelMetadata):
            raise ValueError(
                'channel metadata must be of type ChannelMetadata')
        if isinstance(ntv_idxs, (list, np.ndarray)):
            pass
        elif isinstance(ntv_idxs, (int, float)):
            ntv_idxs = [ntv_idxs]
        elif ntv_idxs in [None]:
            ntv_idxs = channel_metadata.index(SignalType.AMP).ntv
        else:
            raise ValueError("ntv_idxs must be scalar or list or None")
        reg_float = [0, 0]
        reg_bool = [False, False]
        if neg_thr not in [None]:
            if neg_thr not in ["on", "off"]:
                raise ValueError("neg_thr must be ``on`` or ``off``")
            reg_float[0] = 1.0 if neg_thr in ["on"] else 0.0
            reg_bool[0] = True
        if pos_thr not in [None]:
            if pos_thr not in ["on", "off"]:
                raise ValueError("pos_thr must be ``on`` or ``off``")
            reg_float[1] = 1.0 if pos_thr in ["on"] else 0.0
            reg_bool[1] = True
        if weak_thr in [False]:
            arg = common_pb2.SORTER_THR_ACTIVATE
        elif weak_thr in [True]:
            arg = common_pb2.SORTER_WEAK_THR_ACTIVATE
        else:
            raise ValueError("weak_thr must be True | False")
        req = [
            common_pb2.SpikeSorterSetParamRequest(
                cmd=arg, regFloat64=reg_float, regBool=reg_bool, ntvChanIdx=ntv_idxs
            )
        ]
        if self._is_allego():
            req.spikeSorterID = PRIMARY_CACHE_STREAM_GROUP_ID
            api_allego.set_kpi_params(
                NEURONS1_ADDR, common_pb2.SpikeSorterSetParamsRequest(params=req))
        else:
            req[0].spikeSorterID = dataset_metadata.attributes['dsource_id']
            api_videre.set_kpi_params(self.__parent._server_address(hub_name, RadiensService.CORE),
                                      common_pb2.SpikeSorterSetParamsRequest(params=req))
        return '[{} channels] {}: neg thr={}, pos_thr={}' .format(len(ntv_idxs),
                                                                  'detect' if arg in [common_pb2.SORTER_THR_LEVEL,
                                                                                      common_pb2.SORTER_THR_LEVEL_SD] else 'weak',
                                                                  '{}'.format(neg_thr) if reg_bool[0] in [
            True] else 'n/a',
            '{}'.format(pos_thr) if reg_bool[1] in [True] else 'n/a')

    def set_event_window(self, pre_ms: float, post_ms: float, dataset_metadata=None, hub_name=DEFAULT_HUB_ID, channel_metadata=None) -> str:
        """
        Sets signal metrics event (spike) window

        Parameters:
            pre_ms (float): ms before event
            post_ms (float): ms after event
            dataset_metadata (~radiens.lib.dataset_metadata.DatasetMetadata): dataset metadata object
            channel_metadata (~radiens.lib.channel_metadata.ChannelMetadata): Allego connected channels

        Returns:
            msg (str): summary description of event window

        Example:
            >>> client.signal_metrics().set_event_window(2, 2)
            '[32 channels] event window: pre-threshold=2 ms, post-threshold=2 ms

        See Also:
             :py:meth:`set_threshold_level`
             :py:meth:`get_channel_metadata`

        """
        if not self._is_allego() and not isinstance(dataset_metadata, DatasetMetadata):
            raise ValueError('videre: dataset_metadata must be provided')
        if not isinstance(pre_ms, (int, float)) or not isinstance(post_ms, (int, float)):
            return 'pre_ms and post_ms are required scalars'
        if channel_metadata in [None]:
            channel_metadata = self.__parent.get_channel_metadata(
            ) if self._is_allego() else dataset_metadata.channel_metadata
        if not isinstance(channel_metadata, ChannelMetadata):
            raise ValueError(
                'channel metadata must be of type ChannelMetadata')
        reg_float = [pre_ms, post_ms]
        reg_bool = [True, True]
        req = [common_pb2.SpikeSorterSetParamRequest(cmd=common_pb2.SORTER_THR_WDW,
                                                     regFloat64=reg_float,
                                                     regBool=reg_bool,
                                                     ntvChanIdx=channel_metadata.index(SignalType.AMP).ntv)]
        if self._is_allego():
            req.spikeSorterID = PRIMARY_CACHE_STREAM_GROUP_ID
            api_allego.set_kpi_params(
                NEURONS1_ADDR, common_pb2.SpikeSorterSetParamsRequest(params=req))
        else:
            req[0].spikeSorterID = dataset_metadata.attributes['dsource_id']
            api_videre.set_kpi_params(self.__parent._server_address(hub_name, RadiensService.CORE),
                                      common_pb2.SpikeSorterSetParamsRequest(params=req))
        return '[{} channels] event window: pre-threshold={} ms, post-threshold={} ms' .format(len(channel_metadata.index(SignalType.AMP).ntv),
                                                                                               '{}'.format(
                                                                                                   pre_ms),
                                                                                               '{}'.format(post_ms))

    def set_event_shadow(self, shadow_ms: float, dataset_metadata=None, channel_metadata=None, hub_name=DEFAULT_HUB_ID) -> str:
        """
        Sets signal metrics event (spike) shadow time in milliseconds.

        Parameters:
            shadow_ms (float): duration of shadow in milliseconds
            dataset_metadata (~radiens.lib.dataset_metadata.DatasetMetadata): dataset metadata object
            channel_metadata (~radiens.lib.channel_metadata.ChannelMetadata): Allego connected channels

        Returns:
            msg (str): summary description of event shadow

        Example:
            >>> client.signal_metrics().set_event_shadow(1)
            '[32 channels] event shadow 1 ms'

        See Also:
             :py:meth:`set_threshold_level`
             :py:meth:`get_channel_metadata`

        """
        if self.__parent.type not in ["allego"] and not isinstance(
            dataset_metadata, DatasetMetadata
        ):
            raise ValueError("videre: dataset_metadata must be provided")
        if not isinstance(shadow_ms, (int, float)):
            return 'shadow_ms is a required scalar'
        if channel_metadata in [None]:
            channel_metadata = self.__parent.get_channel_metadata(
            ) if self._is_allego() else dataset_metadata.channel_metadata
        if not isinstance(channel_metadata, ChannelMetadata):
            raise ValueError(
                'channel metadata must be of type ChannelMetadata')
        reg_float = [shadow_ms, np.NaN]
        reg_bool = [True, False]
        req = [common_pb2.SpikeSorterSetParamRequest(cmd=common_pb2.SORTER_SHADOW,
                                                     regFloat64=reg_float,
                                                     regBool=reg_bool,
                                                     ntvChanIdx=channel_metadata.index(SignalType.AMP).ntv)]
        if self._is_allego():
            req.spikeSorterID = PRIMARY_CACHE_STREAM_GROUP_ID
            api_allego.set_kpi_params(
                NEURONS1_ADDR, common_pb2.SpikeSorterSetParamsRequest(params=req))
        else:
            req[0].spikeSorterID = dataset_metadata.attributes['dsource_id']
            api_videre.set_kpi_params(self.__parent._server_address(hub_name, RadiensService.CORE),
                                      common_pb2.SpikeSorterSetParamsRequest(params=req))
        return '[{} channels] event shadow {} ms,' .format(len(channel_metadata.index(SignalType.AMP).ntv),
                                                           '{}'.format(shadow_ms))

    def set_packet_duration(self, packet_dur_sec: float, dataset_metadata=None, hub_name=DEFAULT_HUB_ID) -> None:
        """
        Sets signal metrics packet duration in seconds

        Parameters:
            packet_dur_sec (float): duration in seconds of signals used to calculate signal metrics packet
            dataset_metadata (~radiens.lib.dataset_metadata.DatasetMetadata): dataset metadata object

        Returns:
            msg (str): summary description of threshold levels

        Example:
            >>> client.signal_metrics().set_packet_duration(neg_thr='on')
            None

        See Also:
             :py:meth:`set_threshold_level`
             :py:meth:`get_channel_metadata`

        """
        if self.__parent.type not in ["allego"] and not isinstance(
            dataset_metadata, DatasetMetadata
        ):
            raise ValueError("videre: dataset_metadata must be provided")
        if not isinstance(packet_dur_sec, (int, float)):
            return "packet_dur_sec is a required scalar"
        req = common_pb2.SetKpiParamRequest(param=packet_dur_sec)
        if self._is_allego():
            req.streamGroupId = PRIMARY_CACHE_STREAM_GROUP_ID
            api_allego.set_kpi_packet_dur(NEURONS1_ADDR, req)
        else:
            req.streamGroupId = dataset_metadata.attributes['dsource_id']
            api_videre.set_kpi_packet_dur(
                self.__parent._server_address(hub_name, RadiensService.CORE), req)

    def get_params(self, dataset_metadata: DatasetMetadata = None, hub_name=DEFAULT_HUB_ID) -> pd.DataFrame:
        """
        Returns signal metrics parameters as a table.

        Parameters:
            dataset_metadata (DatasetMetadata): metadata on linked dataset
            hub_name (str): Radiens hub name (default = DEFAULT_HUB, optional)

        Parameters:
            dataset_metadata (~radiens.lib.dataset_metadata.DatasetMetadata): dataset metadata object

        Returns:
            pandas.DataFrame
        """
        if self._is_allego():
            api_allego.get_kpi_params(NEURONS1_ADDR)
        else:
            if not isinstance(dataset_metadata, DatasetMetadata):
                raise ValueError(
                    'videre: dataset_metadata must be a DatasetMetadata')
            return api_videre.get_kpi_params(self.__parent._server_address(hub_name, RadiensService.CORE), dataset_metadata.attributes['dsource_id'])

    def get_metrics_status(self, dataset_metadata: DatasetMetadata = None, hub_name=DEFAULT_HUB_ID) -> SignalMetricsStatus:
        """
        Returns the status of the signal metrics service.  This is an advanced function and is not typically used.

        Parameters:
            dataset_metadata (~radiens.lib.dataset_metadata.DatasetMetadata): dataset metadata object

        Returns:
            status (dict): dict of status parameters, with keys='time_range','packet_dur_sec', 'beta', 'is_tracking_cache', 'wall_time_start', 'persistence_sec'
        """
        if self._is_allego():
            return api_allego.get_kpi_status(KPI_ADDR)
        else:
            if not isinstance(dataset_metadata, DatasetMetadata):
                raise ValueError(
                    'videre: dataset_metadata must be a DatasetMetadata')
            return api_videre.get_kpi_status(self.__parent._server_address(hub_name, RadiensService.CORE), dataset_metadata.attributes['dsource_id'])

    def get_metrics(self,
                    time_range: Union[int, float, list, np.ndarray] = None,
                    metrics: list[MetricID] = None,
                    ntv_idxs=None,
                    tail=True,
                    plot: bool = True,
                    file: str = None,
                    data: bool = True,
                    dataset_metadata=None,
                    hub_name=DEFAULT_HUB_ID,
                    channel_metadata=None) -> SignalMetrics:
        """
        Gets the requested signal metrics from a linked dataset.

        Parameters:
            time_range (int, float, list, ndarray): if parent client is :py:class:`.AllegoClient`, then this parameter can be `float` or `int`; otherwise, it must be `list` or `~numpy.ndarray`
            metrics (list): a list of :py:class:`~radiens.lib.sig_metrics.METRIC_ID`. It specified a default list is used.
            dataset_metadata (~radiens.lib.dataset_metadata.DatasetMetadata): dataset metadata object
            channel_metadata (~radiens.lib.channel_metadata.ChannelMetadata): Allego connected channels


        Returns:
            SignalMetrics (~radiens.lib.sig_metrics.SignalMetrics): Requested metrics

        """
        if self._is_allego():
            if not isinstance(time_range, (float, int)) and time_range not in [None]:
                raise ValueError(
                    'allego: time_range must be None or scalar lag time (sec) from head of primary cache')
            if not isinstance(channel_metadata, (ChannelMetadata)) and channel_metadata not in [None]:
                raise ValueError(
                    'allego: channel metadata must be None or a ChannelMetadata')
            _tr = [5, np.NAN] if time_range in [None] else [time_range, np.NAN]
            channel_metadata = self.__parent.get_channel_metadata() if channel_metadata in [
                None] else channel_metadata
        else:
            if not isinstance(dataset_metadata, DatasetMetadata):
                if dataset_metadata is None:
                    raise ValueError(
                        'videre: dataset_metadata must be provided')
                else:
                    raise ValueError(
                        'videre: dataset metadata must be a DatasetMetadata')

            if not isinstance(time_range, (list, np.ndarray)) and time_range is not None:
                raise ValueError(
                    'videre: time_range must be None or [start, end] (sec) with respect to datasource time range')

            channel_metadata = dataset_metadata.channel_metadata
            _tr = dataset_metadata.time_range.sec if time_range is None else np.array(
                time_range, dtype=np.float64)
        ntv_idxs = channel_metadata.index(SignalType.AMP).ntv
        status = self.get_metrics_status(dataset_metadata=dataset_metadata)

        if metrics in [None]:
            metrics = [MetricID(mode=MetricMode.BASE, name=MetricName.RMS),
                       MetricID(mode=MetricMode.BASE,
                                name=MetricName.NOISE_UV),
                       MetricID(mode=MetricMode.BASE,
                                name=MetricName.MAX_MIN_DIFF_ABS),
                       MetricID(mode=MetricMode.BASE, name=MetricName.SNR),
                       MetricID(mode=MetricMode.BASE,
                                name=MetricName.EVENT_RATE),
                       MetricID(mode=MetricMode.BASE,
                                name=MetricName.EVENT_MEAN_MAX_MIN_DIFF_ABS),
                       ]
        _metrics = []
        for m in metrics:
            _metrics.append(common_pb2.KpiMetricID(
                mode=m.mode.value, name=m.name.value))
        arg = common_pb2.BundleReq(ntvIdxs=ntv_idxs, tR=common_pb2.TimeRange(
            timeRangeSec=list(_tr), fs=status.time_range.fs), metrics=_metrics, isTail=tail)
        req = common_pb2.KpiMetricsReq(stype=common_pb2.PRI,
                                       arg=arg,
                                       isPlot=plot,
                                       path=file,
                                       isReturnData=data)
        if self._is_allego():
            req.streamGroupId = PRIMARY_CACHE_STREAM_GROUP_ID
            return api_allego.get_kpi_metrics(KPI_ADDR, req)
        else:
            if not isinstance(dataset_metadata, DatasetMetadata):
                raise ValueError(
                    'videre: dataset_metadata must be a DatasetMetadata')
            req.streamGroupId = dataset_metadata.attributes['dsource_id']
            return api_videre.get_kpi_metrics(self.__parent._server_address(hub_name, RadiensService.CORE), dataset_metadata.attributes['dsource_id'], req)



from collections.abc import Iterable
from pathlib import Path

import pandas as pd
from radiens.api import api_curate, api_videre
from radiens.api.api_utils.util import BaseClient, to_radiens_file_type
from radiens.grpc_radiens import common_pb2, datasource_pb2
from radiens.lib.dataset_metadata import DatasetMetadata
from radiens.lib.fsys import FileSysResponse
from radiens.utils.constants import DEFAULT_HUB_ID
from radiens.utils.enums import RadiensService


class FileSystemClient(BaseClient):
    """
    Container for Radiens file system commands.
    """

    def __init__(self):
        """ """
        super().__init__()

    @property
    def hubs(self) -> dict:
        """
        dict of active radiens hubs, with hub ID as key.
        """
        return self._hubs()

    @property
    def id(self) -> str:
        """
        UID of this client session.
        """
        return self._id()

    def ls(self, req_dir: any, sort_by="date", hub_name=DEFAULT_HUB_ID) -> FileSysResponse:
        """
        Aliases client.list_dir() to list Radiens recording and/or spikes files that are in the requested directory.

        See Also:
            :py:meth:`list_dir()`
        """
        pArgs = []
        try:
            pArgs = [str(Path(req_dir).expanduser().resolve())]
        except TypeError:
            try:
                for x in req_dir:
                    pArgs.append(str(Path(x).expanduser().resolve()))
            except TypeError:
                raise TypeError(
                    "req_dir must be a str or Path or a list of them")

        return api_curate.dsrc_list_dir(self._server_address(hub_name, RadiensService.CORE), pArgs, sort_by)

    def list_dir(
        self, req_dir: any, sort_by="date", include="all", hub_name=DEFAULT_HUB_ID
    ) -> pd.DataFrame:
        """
        Lists Radiens recording and/or spikes files that are in the requested directory.

        Parameters:
            req_dir (str, pathlib.Path): requested directory
            sort_by (str): flag to specify the ordering of the returned file list, 'date', 'name', 'size', 'type' (default='date)
            include (str): flag to specify file types, 'recording', 'time-series', 'spikes', 'all' (default='all')
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            file_type (pandas.DataFrame): table of files in the requested directory

        Example:
            >>> client.list_dir("./")
            pandas.DataFrame

        Notes:
            Directory name wildcards are allowed, but must resolve to only one directory.

        See Also:
            :py:meth:`ls`
        """
        dsrc_table = self.ls(req_dir, sort_by, hub_name).datasource_table
        if include != "all":
            dsrc_table = dsrc_table[dsrc_table.type == include]
        return dsrc_table

    def copy_files(
        self, source: any, dest: any, force=False, hub_name=DEFAULT_HUB_ID
    ) -> FileSysResponse:
        """
        Copies one or more Radiens files.

        Parameters:
            source (str, pathlib.Path, list): path to source file(s)
            dest (str, pathlib.Path): path to destination file(s)
            force (bool): flag to force copy if destination file already exists (default=False)
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            file_desc (FileSysResponse): details of the command result

        Example:
            >>> client.copy_files("./my_rec_file.xdat", './targ_dir/new_file')
            FileSysResponse

        Notes:
            The destination file type is always the same as the source file type.
            File name wildcards are allowed, but must resolve to only one file.

        See Also:
            :py:meth:`cp`
        """
        return self.cp(source, dest, force, hub_name)

    def cp(
        self, source: any, dest: any, force=False, hub_name=DEFAULT_HUB_ID
    ) -> FileSysResponse:
        """
        Aliases client.copy_files() to copy one or more Radiens files.

        See Also:
            :py:meth:`copy_files`
        """
        src_one = None
        src_many = None
        if isinstance(source, (Path, str)):
            src_one = Path(source).expanduser().resolve()
        elif isinstance(source, (list, Iterable)):
            src_many = []
            for file in source:
                src_many.append(Path(file).expanduser().resolve())
        else:
            raise ValueError("source must be string, Path, or list")

        if not isinstance(dest, (Path, str)):
            raise ValueError("dest must be string or Path")

        dest = Path(dest).expanduser().resolve()
        if src_one is not None:
            src_desc = common_pb2.FileDescriptor(
                path=str(src_one.parent),
                baseName=src_one.stem,
                fileType=to_radiens_file_type(src_one),
            )
            if dest.is_dir():
                dest = Path(dest, src_one.name)
            dest_desc = common_pb2.FileDescriptor(
                path=str(dest.parent), baseName=dest.stem, fileType=to_radiens_file_type(src_one))
            req_one = datasource_pb2.CopyRemoveDataSourceFileRequest.OneFile(
                src=src_desc, dest=dest_desc)
            req = datasource_pb2.CopyRemoveDataSourceFileRequest(
                isForce=bool(force), one=req_one)
            return api_curate.dsrc_copy(self._server_address(hub_name, RadiensService.CORE), req)

        # many files
        if not dest.is_dir():
            raise ValueError("dest must be a path when copying multiple files")
        src_desc = []
        for file in src_many:
            src_desc.append(
                common_pb2.FileDescriptor(
                    path=str(file.parent),
                    baseName=file.stem,
                    fileType=to_radiens_file_type(file),
                )
            )
        req_many = datasource_pb2.CopyRemoveDataSourceFileRequest.MultipleFiles(
            src=src_desc, destPath=str(dest), destFileType=to_radiens_file_type(Path(file)))
        req = datasource_pb2.CopyRemoveDataSourceFileRequest(
            isForce=bool(force), many=req_many)
        return api_curate.dsrc_copy(self._server_address(hub_name, RadiensService.CORE), req)

    def delete_files(self, source: any, hub_name=DEFAULT_HUB_ID) -> FileSysResponse:
        """
        Deletes (aka removes) one or more Radiens files.

        Parameters:
            source (str, pathlib.Path, list): path to source file(s)
            dry_run (bool): flag to list the requested files without deleting them (default=False)
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            file_type (pandas.DataFrame): table of copied files.

        Example:
            >>> client.delete_files("./my_rec_file.xdat")
            pandas.DataFrame

        Notes:
            The requested files are permanently deleted from the file system.
            File name wildcards are allowed, but must resolve to only one file.

        See Also:
            :py:meth:`rm`
        """
        return self.rm(source, hub_name)

    def rm(self, source: any, hub_name=DEFAULT_HUB_ID) -> FileSysResponse:
        """
        Aliases client.delete_files() to delete (aka remove) one or more Radiens files.

        See Also:
            :py:meth:`delete_files`
        """
        src_many = []
        if isinstance(source, (Path, str)):
            src_many = [Path(source).expanduser().resolve()]
        elif isinstance(source, (list, Iterable)):
            for file in source:
                src_many.append(Path(file).expanduser().resolve())
        else:
            raise ValueError("source type must be string, Path, or list")

        src_desc = []
        for file in src_many:
            src_desc.append(common_pb2.FileDescriptor(
                path=str(file.parent), baseName=file.stem, fileType=to_radiens_file_type(file)))
        req_many = datasource_pb2.CopyRemoveDataSourceFileRequest.MultipleFiles(
            src=src_desc)
        req = datasource_pb2.CopyRemoveDataSourceFileRequest(
            isForce=True, many=req_many)
        return api_curate.dsrc_remove(self._server_address(hub_name, RadiensService.CORE), req)

    def rename_file(
        self, source: any, dest=any, force=False, validate=True, hub_name=DEFAULT_HUB_ID
    ) -> FileSysResponse:
        """
        Renames (aka moves) one Radiens file.

        Parameters:
            source (str, pathlib.Path): path to source file
            dest (str, pathlib.Path): path to destination file
            dry_run (bool): flag to list the requested files without deleting them (default=False)
            force (bool): flag to force renaming if the destination file exists (default=False)
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            file_type (pandas.DataFrame): table of renamed file.

        Example:
            >>> client.rename_files("./my_rec_file.xdat", "./targ_dir/my_rec_file")
            pandas.DataFrame

        Notes:
            The destination file type is the same as the source file type.
            File name wildcards are allowed, but must resolve to only one file.

        See Also:
            :py:meth:`mv`
        """
        return self.mv(source, dest, force, validate, hub_name)

    def mv(
        self, source: any, dest=any, force=True, validate=True, hub_name=DEFAULT_HUB_ID
    ) -> FileSysResponse:
        """
        Aliases client.rename_file() to rename (aka move) one Radiens file.

        See Also:
            :py:meth:`rename_file`
        """
        if isinstance(source, (Path, str)):
            src = Path(source).expanduser().resolve()
        else:
            raise ValueError("source must be string or Path")
        if isinstance(dest, (Path, str)):
            dest = Path(dest).expanduser().resolve()
        else:
            raise ValueError("dest must be string or Path")

        src_desc = common_pb2.FileDescriptor(
            path=str(src.parent), baseName=src.stem, fileType=to_radiens_file_type(src)
        )
        dest_desc = common_pb2.FileDescriptor(
            path=str(dest.parent), baseName=dest.stem, fileType=to_radiens_file_type(dest))
        req = datasource_pb2.MoveDataSourceFileRequest(
            isForce=force, isValidate=validate, src=src_desc, dest=dest_desc)
        return api_curate.dsrc_move(self._server_address(hub_name, RadiensService.CORE), req)

    def link_data_file(
        self, source: any, calc_metrics=False, hub_name=DEFAULT_HUB_ID
    ) -> DatasetMetadata:
        """
        Links a Radiens data file to the Radiens hub as a new dataset.

        Parameters:
            source (str, pathlib.Path): path to source file
            calc_metrics (bool) optional: set True to calculate signal metrics (default=False)
            hub_name (str) optional: radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            data_source (DatasetMetadata): data file metadata

        See Also:
            :py:meth:`rename_file`
            :py:meth:`_clear_dataset()`
            :py:meth:`get_dataset_ids()`
        """
        if isinstance(source, (Path, str)):
            source = Path(source).expanduser().resolve()
        else:
            raise ValueError("source must be string or Path")
        req = datasource_pb2.DataSourceSetSaveRequest(
            path=str(source.parent),
            baseName=source.stem,
            fileType=to_radiens_file_type(source),
            dsourceID=source.stem + self.id,
            isBackgroundKPI=calc_metrics,
        )
        return api_videre.set_datasource(
            self._server_address(hub_name, RadiensService.CORE), req
        )

    def get_data_file_metadata(
        self, source: any, hub_name=DEFAULT_HUB_ID
    ) -> DatasetMetadata:
        """
        Returns the meta data for a Radiens data file.

        Parameters:
            source (str, pathlib.Path): path to source file
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            data_source (DatasetMetadata): data file metadata

        See Also:
            :py:meth:`rename_file`
            :py:meth:`_clear_dataset()`
            :py:meth:`get_dataset_ids()`
        """
        if isinstance(source, (Path, str)):
            source = Path(source).expanduser().resolve()
        else:
            raise ValueError('source must be string or Path')
        req = datasource_pb2.DataSourceSetSaveRequest(path=str(source.parent), baseName=source.stem, fileType=to_radiens_file_type(
            source), dsourceID=source.stem+self.id, isBackgroundKPI=False)
        dsrc = api_videre.set_datasource(
            self._server_address(hub_name, RadiensService.CORE), req)
        self._clear_dataset(dsrc.id)
        dsrc.clear_dataset_id()
        return dsrc

    def _clear_dataset(self, source=[], dataset_id=[], hub_name=DEFAULT_HUB_ID) -> str:
        if not isinstance(dataset_id, (list, str, Iterable)):
            raise ValueError("dataset_id must be string or list of strings")
        if isinstance(dataset_id, str):
            dataset_id = [dataset_id]
        if isinstance(source, (str, Path)):
            source = [Path(source)]
        _get_ids = []
        for _id in self._get_dataset_ids(hub_name):
            if len(dataset_id) > 0 and dataset_id[0] == "all" and _id.find(self.id) > 0:
                _get_ids.append(_id)
            else:
                for req_id in dataset_id:
                    if req_id == _id:
                        _get_ids.append(_id)
            for src in source:
                if _id.find(Path(src).stem) > 0 and _id.find(self.id) > 0:
                    _get_ids.append(_id)
        return api_videre.unlink_datasource(
            self._server_address(hub_name, RadiensService.CORE), _get_ids
        )

    def _get_dataset_ids(self, hub_name=DEFAULT_HUB_ID) -> list:
        return api_videre.list_datasource_ids(
            self._server_address(hub_name, RadiensService.CORE)
        )



import unittest

import radiens.utils.config as config


class base_utest(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_metadata_constructor(self):
        pass


import os
import socket
import subprocess
import sys
import uuid
from pathlib import Path
from typing import Union

import grpc
import numpy as np
import pandas as pd
import radiens.utils.config as cfg
from radiens.grpc_radiens import allegoserver_pb2, common_pb2
from radiens.utils.constants import DEFAULT_HUB_ID, DEFAULT_IP
from radiens.utils.enums import RadiensService


def parse_stim_param_resp(resp: allegoserver_pb2.StimParamsReply):
    resp_dict = dict()
    for stim_sys_chan_idx, params in resp.params.items():
        resp_dict[stim_sys_chan_idx] = {}
        resp_dict[stim_sys_chan_idx]['stim_shape'] = params.stimShape
        resp_dict[stim_sys_chan_idx]['stim_polarity'] = params.stimPolarity
        resp_dict[stim_sys_chan_idx]['first_phase_duration_us'] = params.firstPhaseDuration
        resp_dict[stim_sys_chan_idx]['second_phase_duration_us'] = params.secondPhaseDuration
        resp_dict[stim_sys_chan_idx]['interphase_delay_us'] = params.interphaseDelay
        resp_dict[stim_sys_chan_idx]['first_phase_amplitude_uA'] = params.firstPhaseAmplitude
        resp_dict[stim_sys_chan_idx]['second_phase_amplitude_uA'] = params.secondPhaseAmplitude
        resp_dict[stim_sys_chan_idx]['baseline_voltage'] = params.baselineVoltage
        resp_dict[stim_sys_chan_idx]['trigger_edge_or_level'] = params.triggerEdgeOrLevel
        resp_dict[stim_sys_chan_idx]['trigger_high_or_low'] = params.triggerHighOrLow
        resp_dict[stim_sys_chan_idx]['enabled'] = params.enabled
        resp_dict[stim_sys_chan_idx]['post_trigger_delay_us'] = params.postTriggerDelay
        resp_dict[stim_sys_chan_idx]['pulse_or_train'] = params.pulseOrTrain
        resp_dict[stim_sys_chan_idx]['number_of_stim_pulses'] = params.numberOfStimPulses
        resp_dict[stim_sys_chan_idx]['pulse_train_period_us'] = params.pulseTrainPeriod
        resp_dict[stim_sys_chan_idx]['refactory_period_ms'] = params.refractoryPeriod
        resp_dict[stim_sys_chan_idx]['pre_stim_amp_settle_us'] = params.preStimAmpSettle
        resp_dict[stim_sys_chan_idx]['post_stim_amp_settle_us'] = params.postStimAmpSettle
        resp_dict[stim_sys_chan_idx]['maintain_amp_settle_us'] = params.maintainAmpSettle
        resp_dict[stim_sys_chan_idx]['enable_amp_settle'] = params.enableAmpSettle
        resp_dict[stim_sys_chan_idx]['post_stim_charge_recovery_on_us'] = params.postStimChargeRecovOn
        resp_dict[stim_sys_chan_idx]['post_stim_charge_recovery_off_us'] = params.postStimChargeRecovOff
        resp_dict[stim_sys_chan_idx]['enable_charge_recovery'] = params.enableChargeRecovery
        resp_dict[stim_sys_chan_idx]['trigger_source_is_keypress'] = params.triggerSourceIsKeypress
        resp_dict[stim_sys_chan_idx]['trigger_source_idx'] = params.triggerSourceIdx
        resp_dict[stim_sys_chan_idx]['stim_sys_chan_idx'] = params.stimSysChanIdx

    return resp_dict


def stim_params_dict_to_req(req_dict: dict):
    return allegoserver_pb2.StimParams(
        baselineVoltage=req_dict['baseline_voltage'],
        firstPhaseDuration=req_dict['first_phase_duration_us'],
        enabled=req_dict['enabled'],
        enableAmpSettle=req_dict['enable_amp_settle'],
        enableChargeRecovery=req_dict['enable_charge_recovery'],
        firstPhaseAmplitude=req_dict['first_phase_amplitude_uA'],
        interphaseDelay=req_dict['interphase_delay_us'],
        maintainAmpSettle=req_dict['maintain_amp_settle_us'],
        numberOfStimPulses=req_dict['number_of_stim_pulses'],
        postStimAmpSettle=req_dict['post_stim_amp_settle_us'],
        postStimChargeRecovOff=req_dict['post_stim_charge_recovery_off_us'],
        postStimChargeRecovOn=req_dict['post_stim_charge_recovery_on_us'],
        postTriggerDelay=req_dict['post_trigger_delay_us'],
        preStimAmpSettle=req_dict['pre_stim_amp_settle_us'],
        pulseOrTrain=req_dict['pulse_or_train'],
        pulseTrainPeriod=req_dict['pulse_train_period_us'],
        refractoryPeriod=req_dict['refactory_period_ms'],
        secondPhaseAmplitude=req_dict['second_phase_amplitude_uA'],
        secondPhaseDuration=req_dict['second_phase_duration_us'],
        stimPolarity=req_dict['stim_polarity'],
        stimShape=req_dict['stim_shape'],
        stimSysChanIdx=req_dict['stim_sys_chan_idx'],
        triggerEdgeOrLevel=req_dict['trigger_edge_or_level'],
        triggerHighOrLow=req_dict['trigger_high_or_low'],
        triggerSourceIsKeypress=req_dict['trigger_source_is_keypress'],
        triggerSourceIdx=req_dict['trigger_source_idx'],
    )


def to_radiens_file_type(p: Path):
    ext = p.suffix.lower()
    if ext in ['.rhd']:
        return common_pb2.RHD
    if ext in ['.xdat', '']:
        return common_pb2.XDAT
    if ext in ['.csv']:
        return common_pb2.CSV
    if ext in ['.hdf5', '.h5']:
        return common_pb2.HDF5
    if ext in ['.nex5']:
        return common_pb2.NEX5
    if ext in ['.nwb']:
        return common_pb2.NWB
    if ext in ['.kilosort2']:
        return common_pb2.KILOSORT2
    if ext in ['.nsx']:
        return common_pb2.NSX
    if ext in ['.tdt']:
        return common_pb2.TDT
    if ext in ['.spikes']:
        return common_pb2.SPIKES
    raise ValueError('{} is invalid or unknown file extension'.format(ext))


def to_file_ext(file_type: common_pb2.RadixFileTypes):
    return common_pb2.RadixFileTypes.Name(file_type).lower()


def path_to_source_sink_transform(p: Path) -> common_pb2.DataSourceSinkTransformParams:
    return common_pb2.DataSourceSinkTransformParams(dsrcName=p.expanduser().resolve().stem, path=str(p.expanduser().resolve().parent), fileType=to_radiens_file_type(p))


def source_sink_transform_to_path(pb_node: common_pb2.DataSourceSinkTransformParams) -> Path:
    return Path(pb_node.datasourceSinkParams.path, pb_node.datasourceSinkParams.dsrcName +
                '.{}'.format(to_file_ext(pb_node.datasourceSinkParams.fileType)))


def to_matrix_from_protobuf_radix_matrix(arg: common_pb2.RadixMatrixBytes) -> np.ndarray:
    return np.reshape(np.frombuffer(arg.data, dtype=np.float32), (arg.shape[0], arg.shape[1])).astype(np.float64)


def to_matrix_from_protobuf_dense_matrix(arg: common_pb2.DenseMatrix) -> np.ndarray:
    return np.reshape(np.frombuffer(arg.data, dtype=np.float64), (arg.rows, arg.cols)).astype(np.float64)


class BaseClient:
    """
    """

    def __init__(self):
        self._client_id = ''  # '_-_'+str(uuid.uuid4())[4:13]
        self._base_hubs = {DEFAULT_HUB_ID: {
            'ip_address': DEFAULT_IP,
            'ports': cfg.get_radiens_port_dict(),
        }}

    def _server_address(self, hub_id, service: RadiensService):
        # creates server address from hub and service
        try:
            return '{}:{}'.format(self._hubs()[hub_id]['ip_address'], self._hubs()[hub_id]['ports'][service])
        except KeyError:
            raise AssertionError(
                'invalid service name = {}'.format(service.name))

    def _hubs(self) -> dict:
        return self._base_hubs

    def _id(self) -> str:
        return self._client_id


def get_dataset_id(
        df_avail: pd.DataFrame,
        self_id: str,
        hub: str,
        dataset_idx: int = None,
        dataset_id: str = None,
        path: Union[Path, str] = None,
        fail_hard: bool = False) -> any:
    if dataset_idx is None and dataset_id is None:
        if isinstance(path, (str, Path)):
            try:
                return df_avail.loc[df_avail['dataset ID'] == Path(path).stem + self_id]['dataset ID'].values[0], Path(path)
            except Exception:
                if fail_hard:
                    raise ValueError(
                        '{} is not a linked data file on hub {}'.format(Path(path), hub))
                print('no op, {} is not a linked data file on hub {}'.format(
                    Path(path), hub))
                return None, None
        else:
            raise ValueError(
                'path must be a string or Path when dataset_idx and dataset_id are both None')
    if dataset_idx is not None:
        try:
            return df_avail.loc[[int(dataset_idx)]]['dataset ID'].values[0], None
        except Exception as ex:
            if fail_hard:
                raise ValueError(
                    '{} is not an available dataset index on hub {}'.format(dataset_idx, hub))
            print('no op, {} is not an available dataset index on hub {}'.format(
                dataset_idx, hub))
            return None, None
    if dataset_id is not None:
        try:
            return df_avail.loc[df_avail['dataset ID'] == dataset_id]['dataset ID'].values[0], None
        except Exception as ex:
            if fail_hard:
                raise ValueError(
                    'no op, {} is not an available dataset ID on hub {}'.format(dataset_id, hub))
            print('no op, {} is not an available dataset ID on hub {}'.format(
                dataset_id, hub))
            return None, None
    raise ValueError(
        'dataset index, dataset ID, and path arguments are all invalid')


def get_dataset_idx_from_id(df_avail: pd.DataFrame, dataset_id: str) -> str:
    return df_avail.loc[df_avail['dataset ID'] == dataset_id]['index'].values[0]


def scan_for_open_port() -> int:
    host = "localhost"
    host_ip = socket.gethostbyname(host)
    for port in range(50000, 65535):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect((host_ip, port))
        except socket.error:
            return port
        else:
            s.close()
    return None


def launch_server() -> int:
    port = scan_for_open_port()
    this_path = Path(os.path.realpath(__file__))
    server_file = Path(this_path.parent, '..', 'api_server.py').resolve()
    if port is None:
        raise RuntimeError('did not find open port')
    try:
        subprocess.Popen([sys.executable, str(
            server_file), str(port)], shell=True)
    except subprocess.OSError:
        raise RuntimeError('failed to start radiens-py process')
    return 'localhost:{}'.format(port)


def server_start_script() -> int:
    return str(Path(Path(os.path.realpath(__file__)).parent, '..', 'api_server.py').resolve())



from __future__ import annotations

import numbers
from collections import namedtuple
from enum import Enum, auto
from typing import Any, Type, TypeVar

from radiens.grpc_radiens import allegoserver_pb2, common_pb2, spikesorter_pb2

MetricID = namedtuple("METRIC_ID", ['mode', 'name'])
"""
Example:
    >>> metric_id = METRIC_ID(mode=METRIC_MODE.BASE, name=METRIC.MEAN)
"""


MetricStats = namedtuple("METRIC_STATS", ['mean', 'sd', 'max', 'min'])

# enums


class ClientType(Enum):
    """TYPE is used to specify the type of client. 
    """
    ALLEGO = 'AllegoClient'
    VIDERE = 'VidereClient'
    CURATE = 'CurateClient'

    @classmethod
    def parse(cls, x: str | ClientType) -> ClientType:
        if isinstance(x, str):
            if x.lower() == 'allego' or x.lower() == 'allegoclient':
                return cls.ALLEGO
            elif x.lower() == 'videre' or x.lower() == 'videreclient':
                return cls.VIDERE
            elif x.lower() == 'curate' or x.lower() == 'curateclient':
                return cls.CURATE
            else:
                raise ValueError(f"Invalid {cls.__name__} value: {x}")
        elif isinstance(x, cls):
            return x
        else:
            raise TypeError(
                f"Expected str or {cls.__name__} instance, got {x.__class__}.")

    def is_allego(self):
        return self.value == ClientType.ALLEGO.value

    def is_videre(self):
        return self.value == ClientType.VIDERE.value

    def is_curate(self):
        return self.value == ClientType.CURATE.value


class RadiensService(Enum):
    """
    An enumeration of Radiens services
    """

    CORE = auto()
    DASH = auto()
    SORTER = auto()
    DEV = auto()
    RADIENSPY = auto()


class KeyIndex(Enum):
    """
    An enumeration of key index types 

    Example: 
        >>> key_ind = KeyIndex.NTV 

    Valid values are as follows:
    """

    NTV = 0  #: native
    DATA = 1  #: data
    SYS = 2  #: system


# protobuf enums
T = TypeVar('T', bound='GrpcEnum')


class GrpcEnum(Enum):
    """
    A base class for all enumerations used in the gRPC API
    """
    @classmethod
    def all(cls: Type[T]) -> list[T]:
        """
        Returns a list of all enum values

        Returns:
            list[T]: A list of all enum values

        """
        return list(cls)

    @classmethod
    def all_string(cls: Type[T]) -> list[str]:
        """
        Returns a list of string names of the enum values

        Returns:
            list[str]: A list of string names of the enum values

        """
        return [e.name for e in cls]

    @classmethod
    def all_enum(cls: Type[T]) -> list[int]:
        """
        Returns a list of protobuf enum values

        Returns:
            list[int]: A list of protobuf enum values

        """
        return [e.pb_value for e in cls]

    @classmethod
    def _from_enum(cls: Type[T], pb: int) -> T:
        """
        Converts a protobuf enum value to an enum value

        Args:
            pb (int): The protobuf enum value

        Returns:
            T: The enum value

        Raises:
            ValueError: If the protobuf enum value is invalid

        """
        try:
            return cls(pb)
        except ValueError:
            raise ValueError(f"Invalid {cls.__name__} value: {pb}")

    @classmethod
    def _from_string(cls: Type[T], s: str) -> T:
        """
        Converts a string to an enum value

        Args:
            s (str): The string to convert to an enum value (case-insensitive)

        Returns:
            T: The enum value

        Raises:
            ValueError: If the string is invalid
        """
        try:
            return cls[s.upper()]
        except KeyError:
            raise ValueError(f"Invalid {cls.__name__} value: {s}")

    @classmethod
    def parse(cls: Type[T], x: str | int | T) -> T:
        """
        Converts any value to an enum value

        Args:
            x (Any): The value to convert to an enum value

        Returns:
            T: The enum value

        Raises:
            ValueError: If the value is invalid
            TypeError: If the value is not an int, str, or enum instance

        """
        if isinstance(x, numbers.Number):  # includes bool as well
            return cls._from_enum(x)
        elif isinstance(x, str):
            return cls._from_string(x)
        elif isinstance(x, cls):
            return x
        else:
            raise TypeError(
                f"Expected int, str, or {cls.__name__} instance, got {x.__class__}.")

    @classmethod
    def map_into_keys(cls: Type[T], d: dict[T | str | int]) -> dict[T]:
        """
        Parses each key and returns a new dictionary with the enums (instances) as keys. If they keys are already enum instances, they are left unchanged.
        """
        return {cls.parse(k): v for k, v in d.items()}

    @classmethod
    def map_into_keys_as_names(cls: Type[T], d: dict[T | str | int]) -> dict[str]:
        """
        Parses each key and returns a new dictionary with the enum names (i.e., strings) as keys. If the keys are already strings, they are left unchanged.
        """
        return {cls.parse(k).name: v for k, v in d.items()}

    @classmethod
    def map_into_keys_as_values(cls: Type[T], d: dict[T | str | int]) -> dict[int]:
        """
       Parses each key and returns a new dictionary with the enum values (i.e., ints) as keys. If the keys are already ints, they are left unchanged.
        """
        return {cls.parse(k).value: v for k, v in d.items()}

    @property
    def pb_value(self) -> int:
        """
        Returns the protobuf enum value
        """
        return self.value


class RasterMode(GrpcEnum):
    """
    An enumeration of raster modes
    """
    CHANNELS = spikesorter_pb2.RasterMode.CHANNELS
    NEURONS = spikesorter_pb2.RasterMode.NEURONS


class StreamMode(GrpcEnum):
    """
    An enumeration of stream modes
    """
    OFF = allegoserver_pb2.StreamMode.S_OFF
    ON = allegoserver_pb2.StreamMode.S_ON


class RecordMode(GrpcEnum):
    """
    An enumeration of record modes
    """
    OFF = allegoserver_pb2.RecordMode.R_OFF
    ON = allegoserver_pb2.RecordMode.R_ON


class DioMode(GrpcEnum):
    MANUAL = allegoserver_pb2.DIOMode.manual
    EVENTS = allegoserver_pb2.DIOMode.events
    GATED = allegoserver_pb2.DIOMode.gated


class WorkspaceApp(GrpcEnum):
    """
    An enumeration of workspace applications
    """

    ALLEGO = common_pb2.WorkspaceApp.Allego
    CURATE = common_pb2.WorkspaceApp.Curate
    VIDERE = common_pb2.WorkspaceApp.Videre


class PsdScaling(GrpcEnum):
    '''PsdScaling is used to set the power spectral density (PSD) scale. 
    '''
    SPECTRUM = common_pb2.PSDScaling.PSD_SPECTRUM
    DENSITY = common_pb2.PSDScaling.PSD_DENSITY


class FftWindow(GrpcEnum):
    '''FftWindow is used to set the time-domain window function for power spectral density (PSD) analysis.  
    '''
    HAMMING_p01 = common_pb2.PSDWindowType.HAMMING_p01
    HAMMING_p05 = common_pb2.PSDWindowType.HAMMING_p05
    PASS_THROUGH = common_pb2.PSDWindowType.PASS_THROUGH


class StatsVector(GrpcEnum):
    '''Enum for summary statistic vectors'''

    STAT_MEAN = common_pb2.SummaryStatsEnum.SS_STAT_MEAN
    STAT_SD = common_pb2.SummaryStatsEnum.SS_STAT_SD
    STAT_MODE = common_pb2.SummaryStatsEnum.SS_STAT_MODE
    STAT_MIN = common_pb2.SummaryStatsEnum.SS_STAT_MIN
    STAT_MAX = common_pb2.SummaryStatsEnum.SS_STAT_MAX
    STAT_MODE_COUNT = common_pb2.SummaryStatsEnum.SS_STAT_MODE_COUNT
    STAT_MEDIAN = common_pb2.SummaryStatsEnum.SS_STAT_MEDIAN
    STAT_Q25 = common_pb2.SummaryStatsEnum.SS_STAT_Q25
    STAT_Q75 = common_pb2.SummaryStatsEnum.SS_STAT_Q75
    STAT_SKEW = common_pb2.SummaryStatsEnum.SS_STAT_SKEW
    STAT_KURTOSIS = common_pb2.SummaryStatsEnum.SS_STAT_KURTOSIS
    STAT_N = common_pb2.SummaryStatsEnum.SS_STAT_N


class SummaryStats(GrpcEnum):
    '''Enum for spikes (aka 'neural interface') summary statistics'''

    MIN = common_pb2.SummaryStatsEnum.SS_STAT_MIN
    MAX = common_pb2.SummaryStatsEnum.SS_STAT_MAX
    MEAN = common_pb2.SummaryStatsEnum.SS_STAT_MEAN
    SD = common_pb2.SummaryStatsEnum.SS_STAT_SD
    MODE = common_pb2.SummaryStatsEnum.SS_STAT_MODE
    COUNT = common_pb2.SummaryStatsEnum.SS_STAT_MODE_COUNT
    MEDIAN = common_pb2.SummaryStatsEnum.SS_STAT_MEDIAN
    Q25 = common_pb2.SummaryStatsEnum.SS_STAT_Q25
    Q75 = common_pb2.SummaryStatsEnum.SS_STAT_Q75
    SKEW = common_pb2.SummaryStatsEnum.SS_STAT_SKEW
    KURTOSIS = common_pb2.SummaryStatsEnum.SS_STAT_KURTOSIS
    N = common_pb2.SummaryStatsEnum.SS_STAT_N


class TrsMode(GrpcEnum):
    """
    TrsMode is the time range selector (TRS) enum used to control time range selection in calls to get signals, spikes, power spectral density, or similar data sets.  

    Example:
        >>> sel_mode = TrsMode.SUBSET 

    Given a time range [start, end) in seconds (dataset time), the selector modes are:
    """

    SUBSET = common_pb2.TimeRangeSelMode.TRS_SUBSET  #: selects [start, end)
    #: selects [start, head of stream/cache/file]
    TO_HEAD = common_pb2.TimeRangeSelMode.TRS_TO_HEAD
    #: selects [(head of stream/cache/file - start)
    FROM_HEAD = common_pb2.TimeRangeSelMode.TRS_FROM_HEAD


class MetricMode(GrpcEnum):
    BASE = common_pb2.KpiMetricsMode.Kpi_MetricsMode_Base  # : base packet
    #: running average of all base packets in the stream
    AVG = common_pb2.KpiMetricsMode.Kpi_MetricsMode_Avg
    #: cumulative over all base packets in the stream
    STREAM = common_pb2.KpiMetricsMode.Kpi_MetricsMode_Stream


class MetricName(GrpcEnum):
    #: Number of points in packet
    NUM_PTS = common_pb2.KpiMetricsEnum.KPI_NUM_PTS
    #: Number of interspike intervals in packet
    NUM_PTS_ISI = common_pb2.KpiMetricsEnum.KPI_NUM_PTS_ISI
    #: Packet duration in seconds
    DUR_SEC = common_pb2.KpiMetricsEnum.KPI_DUR_SEC
    #: Mean signal level in packet
    MEAN = common_pb2.KpiMetricsEnum.KPI_MEAN
    #: Minimum signal level in packet
    MIN = common_pb2.KpiMetricsEnum.KPI_MIN
    #: Minimum interspike signal level in packet
    MIN_ISI = common_pb2.KpiMetricsEnum.KPI_MIN_ISI
    #: Maximum signal level in packet
    MAX = common_pb2.KpiMetricsEnum.KPI_MAX
    #: Maximum interspike signal level in packet
    MAX_ISI = common_pb2.KpiMetricsEnum.KPI_MAX_ISI
    #: Absolute maximum signal level in packet
    MAX_ABS = common_pb2.KpiMetricsEnum.KPI_MAX_ABS
    #: Absolute maximum interspike signal level in packet
    MAX_ABS_ISI = common_pb2.KpiMetricsEnum.KPI_MAX_ABS_ISI
    #: Timestamp of minimum signal level in packet
    TIMESTAMP_MIN = common_pb2.KpiMetricsEnum.KPI_TIMESTAMP_MIN
    #: Timestamp of maximum signal level in packet
    TIMESTAMP_MAX = common_pb2.KpiMetricsEnum.KPI_TIMESTAMP_MAX
    #: Absolute difference between maximum and minimum signal levels in packet
    MAX_MIN_DIFF_ABS = common_pb2.KpiMetricsEnum.KPI_MAX_MIN_DIFF_ABS
    #: Absolute difference between maximum and minimum interspike signal levels in packet
    MAX_MIN_DIFF_ABS_ISI = common_pb2.KpiMetricsEnum.KPI_MAX_MIN_DIFF_ABS_ISI
    #: Signal standard deviation in packet
    SD = common_pb2.KpiMetricsEnum.KPI_SD
    #: Signal standard deviation calculated over interspike intervals
    SD_ISI = common_pb2.KpiMetricsEnum.KPI_SD_ISI
    #: Signal variance in packet
    VAR = common_pb2.KpiMetricsEnum.KPI_VAR
    #: Signal variance calculated over interspike intervals
    VAR_ISI = common_pb2.KpiMetricsEnum.KPI_VAR_ISI
    #: Root mean square (RMS) signal value in packet
    RMS = common_pb2.KpiMetricsEnum.KPI_RMS
    #: RMS signal value calculated over interspike intervals
    RMS_ISI = common_pb2.KpiMetricsEnum.KPI_RMS_ISI
    #: Noise level in microvolts
    NOISE_UV = common_pb2.KpiMetricsEnum.KPI_NOISE_UV
    #: Signal-to-noise ratio (SNR) of events in packet
    SNR = common_pb2.KpiMetricsEnum.KPI_SNR
    #: Number of events in packet
    NUM_EVENTS = common_pb2.KpiMetricsEnum.KPI_NUM_EVENTS
    #: Event rate in packet
    EVENT_RATE = common_pb2.KpiMetricsEnum.KPI_EVENT_RATE
    #: Maximum amplitude of events in packet
    EVENT_MAX = common_pb2.KpiMetricsEnum.KPI_EVENT_MAX
    #: Minimum amplitude of events in packet
    EVENT_MIN = common_pb2.KpiMetricsEnum.KPI_EVENT_MIN
    #: Absolute maximum amplitude of events in packet
    EVENT_MAX_ABS = common_pb2.KpiMetricsEnum.KPI_EVENT_MAX_ABS
    #: Absolute difference between maximum and minimum amplitudes of events in packet
    EVENT_MAX_MIN_DIFF_ABS = common_pb2.KpiMetricsEnum.KPI_EVENT_MAX_MIN_DIFF_ABS
    #: Timestamp of minimum amplitude of events in packet
    EVENT_TIMESTAMP_MIN = common_pb2.KpiMetricsEnum.KPI_EVENT_TIMESTAMP_MIN
    #: Timestamp of maximum amplitude of events in packet
    EVENT_TIMESTAMP_MAX = common_pb2.KpiMetricsEnum.KPI_EVENT_TIMESTAMP_MAX
    #: Timestamp of maximum absolute amplitude of events in packet
    EVENT_TIMESTAMP_MAX_ABS = common_pb2.KpiMetricsEnum.KPI_EVENT_TIMESTAMP_MAX_ABS
    #: Timestamp of absolute difference between maximum and minimum amplitudes of events in packet
    EVENT_TIMESTAMP_MAX_MIN_DIFF_ABS = common_pb2.KpiMetricsEnum.KPI_EVENT_TIMESTAMP_MAX_MIN_DIFF_ABS
    #: Mean peak maximum amplitude of events in packet (max over all base packets)
    MEAN_MAX = common_pb2.KpiMetricsEnum.KPI_MEAN_MAX
    #: Mean peak minimum amplitude of events in packet (min over all base packets)
    MEAN_MIN = common_pb2.KpiMetricsEnum.KPI_MEAN_MIN
    #: Mean peak absolute maximum amplitude of events in packet (max over all base packets)
    MEAN_MAX_ABS = common_pb2.KpiMetricsEnum.KPI_MEAN_MAX_ABS
    #: Mean peak absolute amplitude of events in packet (max over all base packets)
    EVENT_MEAN_MAX_MIN_DIFF_ABS = common_pb2.KpiMetricsEnum.KPI_EVENT_MEAN_MAX_MIN_DIFF_ABS
    #: Absolute difference between maximum and minimum amplitudes of events in packet, amplified
    MAX_MIN_DIFF_ABS_AMPLIFIED = common_pb2.KpiMetricsEnum.KPI_MAX_MIN_DIFF_ABS_AMPLIFIED


class Port(GrpcEnum):
    """
    An enumeration of port types

    Example:
        >>> port = Port.A

    Valid values are as follows:
    """
    A = common_pb2.Port.A
    B = common_pb2.Port.B
    C = common_pb2.Port.C
    D = common_pb2.Port.D
    E = common_pb2.Port.E
    F = common_pb2.Port.F
    G = common_pb2.Port.G
    H = common_pb2.Port.H


class SystemMode(GrpcEnum):
    """
    An enumeration of system modes

    Example:
        >>> sys_mode = SystemMode.SMARTBOX_PRO

    Valid values are as follows:
    """

    SMARTBOX_PRO = allegoserver_pb2.BackboneMode.SMARTBOX_PRO
    SMARTBOX_PRO_SINAPS_256 = allegoserver_pb2.BackboneMode.SMARTBOX_PRO_SINAPS_256
    SMARTBOX_PRO_SINAPS_512 = allegoserver_pb2.BackboneMode.SMARTBOX_PRO_SINAPS_512
    SMARTBOX_PRO_SINAPS_1024 = allegoserver_pb2.BackboneMode.SMARTBOX_PRO_SINAPS_1024
    SMARTBOX_CLASSIC = allegoserver_pb2.BackboneMode.SMARTBOX_CLASSIC
    SMARTBOX_SIM_GEN_SINE = allegoserver_pb2.BackboneMode.SMARTBOX_SIM_GEN_SINE
    SMARTBOX_SIM_GEN_SINE_MAPPED = allegoserver_pb2.BackboneMode.SMARTBOX_SIM_GEN_SINE_MAPPED
    SMARTBOX_SIM_GEN_SINE_HIGH_FREQ = allegoserver_pb2.BackboneMode.SMARTBOX_SIM_GEN_SINE_HIGH_FREQ
    SMARTBOX_SIM_GEN_SINE_MULTI_BAND = allegoserver_pb2.BackboneMode.SMARTBOX_SIM_GEN_SINE_MULTI_BAND
    SMARTBOX_SIM_GEN_SPIKES = allegoserver_pb2.BackboneMode.SMARTBOX_SIM_GEN_SPIKES
    OPEN_EPHYS_USB2 = allegoserver_pb2.BackboneMode.OPEN_EPHYS_USB2
    OPEN_EPHYS_USB3 = allegoserver_pb2.BackboneMode.OPEN_EPHYS_USB3
    INTAN_USB2 = allegoserver_pb2.BackboneMode.INTAN_USB2
    INTAN_RECORDING_CONTROLLER_1024 = allegoserver_pb2.BackboneMode.INTAN_RECORDING_CONTROLLER_1024
    INTAN_RECORDING_CONTROLLER_512 = allegoserver_pb2.BackboneMode.INTAN_RECORDING_CONTROLLER_512
    XDAQ_ONE_REC = allegoserver_pb2.BackboneMode.XDAQ_ONE_REC
    XDAQ_ONE_STIM = allegoserver_pb2.BackboneMode.XDAQ_ONE_STIM
    XDAQ_CORE_REC = allegoserver_pb2.BackboneMode.XDAQ_CORE_REC
    XDAQ_CORE_STIM = allegoserver_pb2.BackboneMode.XDAQ_CORE_STIM


class SignalType(GrpcEnum):
    """
    An enumeration of signal types

    Example:
        >>> sig_type = SignalType.AMP

    Valid values are as follows:
    """

    AMP = common_pb2.SignalType.PRI  #: primary signals
    AIN = common_pb2.SignalType.AUX  #: analog in
    DIN = common_pb2.SignalType.DIN  #: digital in
    DOUT = common_pb2.SignalType.DOUT  #: digital out



from collections import namedtuple
from pathlib import Path
from pprint import pprint

import numpy as np
import pandas as pd
from radiens.grpc_radiens import (allegoserver_pb2, common_pb2, datasource_pb2,
                                  spikesorter_pb2)
from radiens.lib.dataset_metadata import FileSetDescriptor

SORTER_STATE = namedtuple("SorterState", ['mode', 'msg', 'time_initialize',
                                          'time_start', 'time_stop', 'launch_time', 'frac_complete', 'is_on',
                                          'kernal_err_msg', 'kernal_warn_msg'])


class SorterState():
    '''
    Radiens container for the spike sorter dashboard
    '''

    def __init__(self, resp):
        """
        """
        if not isinstance(resp, spikesorter_pb2.SpikeSorterDashboardGeneralRec):
            raise ValueError('spike sorter SorterState argument is wrong type')

        self._df_state = pd.DataFrame({'mode': [spikesorter_pb2.SpikeSorterStateSystem.Name(resp.sys)],
                                       'frac_complete': [resp.fracComplete],
                                       'message': [resp.msg],
                                       'init_time': [resp.initializeTime],
                                       'launch_time': [resp.launchTime],
                                       'session_start_time': [resp.sessionStartTime],
                                       'session_stop_time': [resp.sessionStopTime],
                                       'is_on': [resp.isOn],
                                       'kernel_err_message': [resp.kernelErrorMessage],
                                       'kernel_warn_message': [resp.kernelWarnMessage],
                                       }).T

    @ property
    def state(self) -> pd.DataFrame:
        """
        Spike sorter state as a pandas DataFrame

        Access fields via:
            ss.state.loc['mode'][0]
        """
        return self._df_state


class Dashboard():
    '''
    Radiens container for the spike sorter dashboard
    '''

    def __init__(self, resp):
        """
        """
        if not isinstance(resp, spikesorter_pb2.SpikeSorterGetDashboardReply):
            raise ValueError('spike sorter Dashboard argument is wrong type')

        self._df_state = SorterState(resp.general.state)
        self._gen = pd.DataFrame({'ports': [resp.enabledPorts],
                                  'time_range': [np.array(resp.general.timeRange)],
                                  'num_sites_total': [resp.general.numTotalSites],
                                  'num_sites_enabled': [resp.general.numEnabledSites],
                                  'num_sites_active': [resp.general.numActiveSites],
                                  'num_units': [resp.general.numNeurons],
                                  'probe_yield_mean': [resp.general.probeYield],
                                  'site_yield_mean': [resp.general.siteYield],
                                  'num_spikes_processed': [resp.general.numSpikesProcessed],
                                  'num_spikes_labeled': [resp.general.numSpikesLabeled],
                                  'sort_efficiency': [resp.general.sortEfficiency],
                                  'spikes_file_path': [resp.general.sinkDesc.path],
                                  'spikes_file_base_name': [resp.general.sinkDesc.baseName],
                                  }).T
        _sum_stats = []
        _sum_stats.append(pd.DataFrame({'stat': ['probe-snr', 'probe-noise', 'probe-units'],
                                        'port': ['all', 'all', 'all'],
                                        'N': [resp.siteStats.snr[11], resp.siteStats.noise[11], resp.siteStats.neuronYield[11]],
                                        'mean': [resp.siteStats.snr[0], resp.siteStats.noise[0], resp.siteStats.neuronYield[0]],
                                        'sd': [resp.siteStats.snr[1], resp.siteStats.noise[1], resp.siteStats.neuronYield[1]],
                                        'min': [resp.siteStats.snr[3], resp.siteStats.noise[3], resp.siteStats.neuronYield[3]],
                                        'max': [resp.siteStats.snr[4], resp.siteStats.noise[4], resp.siteStats.neuronYield[4]],
                                        'median': [resp.siteStats.snr[6], resp.siteStats.noise[6], resp.siteStats.neuronYield[6]],
                                        'mode': [np.NaN, np.NaN, resp.siteStats.neuronYield[2]],
                                        'mode_cnt': [np.NaN, np.NaN, resp.siteStats.neuronYield[5]],
                                        'q25': [resp.siteStats.snr[7], resp.siteStats.noise[7], resp.siteStats.neuronYield[7]],
                                        'q75': [resp.siteStats.snr[8], resp.siteStats.noise[8], resp.siteStats.neuronYield[8]],
                                        'skew': [resp.siteStats.snr[9], resp.siteStats.noise[9], resp.siteStats.neuronYield[9]],
                                        'kurtosis': [resp.siteStats.snr[10], resp.siteStats.noise[10], resp.siteStats.neuronYield[10]],
                                        }))

        for k in resp.portStats:
            v = resp.portStats[k]
            snr = v.snr
            noise = v.noise
            neuron_yield = v.neuronYield
            if len(snr) != 12:
                snr = [0 for elem in range(0, 12)]
            if len(neuron_yield) != 12:
                neuron_yield = [0 for elem in range(0, 12)]
            if len(noise) != 12:
                noise = [0.01 for elem in range(0, 12)]
            _sum_stats.append(pd.DataFrame({'stat': ['port-snr', 'port-noise', 'port-units'],
                                            'port': [k, k, k],
                                            'N': [snr[11], noise[11], neuron_yield[11]],
                                            'mean': [snr[0], noise[0], neuron_yield[0]],
                                            'sd': [snr[1], noise[1], neuron_yield[1]],
                                            'min': [snr[3], noise[3], neuron_yield[3]],
                                            'max': [snr[4], noise[4], neuron_yield[4]],
                                            'median': [snr[6], noise[6], neuron_yield[6]],
                                            'mode': [np.NaN, np.NaN, neuron_yield[2]],
                                            'mode_cnt': [np.NaN, np.NaN, neuron_yield[5]],
                                            'q25': [snr[7], noise[7], neuron_yield[7]],
                                            'q75': [snr[8], noise[8], neuron_yield[8]],
                                            'skew': [snr[9], noise[9], neuron_yield[9]],
                                            'kurtosis': [snr[10], noise[10], neuron_yield[10]],
                                            }))
        self._df_sum_stats = pd.concat(_sum_stats, axis=0)

    @ property
    def summary_stats(self) -> pd.DataFrame:
        """
        Summary statistics for all ports and probes as a pandas DataFrame
        """
        return self._df_sum_stats

    @ property
    def general(self) -> pd.DataFrame:
        """
        General summary of spike sorter performance
        """
        return self._gen

    @ property
    def state(self) -> pd.DataFrame:
        """
        Spike sorter state as a pandas DataFrame
        """
        return self._df_state.state



import time
import unittest
from datetime import datetime
from pathlib import Path
from pprint import pprint

from radiens.curate_client import CurateClient
from radiens.file_sys_client import FileSystemClient
from radiens.utils.enums import SignalType
from radiens.videre_client import VidereClient
from tqdm import tqdm


class base_utest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls._dir = Path(__file__).parent

    def setUp(self):
        self.client = CurateClient()
        self.videre_client = VidereClient()
        self.fsys_client = FileSystemClient()
        self.radix_data = '~/radix/data'
        self.data_path = '~/radix/data/radiens_test_data'
        self.dataset_0 = '64chan_1s'
        self.tmp_data_path = '~/radix/data/radiens_test_data/tmp'
        self.tmp_dataset_1 = 'tmp_0'

    def tearDown(self) -> None:
        if Path(self.tmp_data_path, self.tmp_dataset_1+'_data.xdat').expanduser().resolve().exists():
            Path(self.tmp_data_path, self.tmp_dataset_1 +
                 '_data.xdat').expanduser().resolve().unlink()
            Path(self.tmp_data_path, self.tmp_dataset_1 +
                 '_timestamp.xdat').expanduser().resolve().unlink()
            Path(self.tmp_data_path, self.tmp_dataset_1 +
                 '.xdat.json').expanduser().resolve().unlink()

    def _has_test_data(self):
        if not Path(self.data_path).expanduser().resolve().exists() or not Path(self.data_path, self.dataset_0+'_data.xdat').expanduser().resolve().exists():
            return False
        return True

    def _dsp_block_until_done(self, stream_id, msg):
        start_time = datetime.now()
        while True:
            elapsed = datetime.now() - start_time
            if elapsed.total_seconds() > 5:
                self.assertFalse(True, 'dsp timed out: {}'.format(msg))
            if self.client.dsp_progress(stream_id) == 1.0:
                break
            time.sleep(0.1)

    def test_apply_filters(self):
        if not self._has_test_data():
            self.skipTest('~/radix/data/radiens_test_data is not set up')

        self.videre_client.clear_dataset(dataset_id='all')
        self.client.dsp_clean_up()

        stream_id = self.client.dsp_low_pass_filter(200.0, Path(self.data_path, self.dataset_0),
                                                    Path(self.tmp_data_path, self.tmp_dataset_1), force=True)
        self._dsp_block_until_done(stream_id, 'low pass filter')
        self.client.dsp_clean_up()

        stream_id = self.client.dsp_high_pass_filter(1000.0, Path(self.data_path, self.dataset_0),
                                                     Path(self.tmp_data_path, self.tmp_dataset_1), force=True)
        self._dsp_block_until_done(stream_id, 'high pass filter')
        self.client.dsp_clean_up()

        stream_id = self.client.dsp_band_pass_filter(200, 1000.0, Path(self.data_path, self.dataset_0),
                                                     Path(self.tmp_data_path, self.tmp_dataset_1), force=True)
        self._dsp_block_until_done(stream_id, 'band pass filter')
        self.client.dsp_clean_up()

        stream_id = self.client.dsp_band_stop_filter(250, 750.0, Path(self.data_path, self.dataset_0),
                                                     Path(self.tmp_data_path, self.tmp_dataset_1), force=True)
        self._dsp_block_until_done(stream_id, 'band stop filter')
        self.client.dsp_clean_up()

        stream_id = self.client.dsp_notch_filter(60, 1.0, Path(self.data_path, self.dataset_0),
                                                 Path(self.tmp_data_path, self.tmp_dataset_1), force=True)
        self._dsp_block_until_done(stream_id, 'notch filter')

        stream_id = self.client.dsp_paired_diff(4, 6, Path(self.data_path, self.dataset_0),
                                                Path(self.tmp_data_path, self.tmp_dataset_1), force=True)
        self._dsp_block_until_done(stream_id, 'paired differential')

        stream_id = self.client.dsp_virtual_ref(23, Path(self.data_path, self.dataset_0),
                                                Path(self.tmp_data_path, self.tmp_dataset_1), force=True)
        self._dsp_block_until_done(stream_id, 'virtual reference')

        stream_id = self.client.dsp_car(Path(self.data_path, self.dataset_0),
                                        Path(self.tmp_data_path, self.tmp_dataset_1), force=True)
        self._dsp_block_until_done(stream_id, 'CAR')

        stream_id = self.client.dsp_time_decimate(2, Path(self.data_path, self.dataset_0),
                                                  Path(self.tmp_data_path, self.tmp_dataset_1+'.xdat'), force=True)
        self._dsp_block_until_done(stream_id, 'time decimate')
        self.assertEqual(20000/2, self.fsys_client.get_data_file_metadata(
            Path(self.tmp_data_path, self.tmp_dataset_1)).time_range.fs)

        stream_id = self.client.dsp_slice_time(16, 17, Path(self.data_path, self.dataset_0),
                                               Path(self.tmp_data_path, self.tmp_dataset_1+'.xdat'), force=True)
        self._dsp_block_until_done(stream_id, 'slice time')
        self.assertEqual(1.0, self.fsys_client.link_data_file(
            Path(self.tmp_data_path, self.tmp_dataset_1)).time_range.dur_sec)

        stream_id = self.client.dsp_slice_channels([4, 6, 23], Path(self.data_path, self.dataset_0),
                                                   Path(self.tmp_data_path, self.tmp_dataset_1+'.xdat'), force=True)
        self._dsp_block_until_done(stream_id, 'slice channels')
        self.assertEqual(32, self.fsys_client.get_data_file_metadata(
            Path(self.data_path, self.dataset_0)).channel_metadata.num_sigs(SignalType.AMP))
        self.assertEqual(3, self.fsys_client.get_data_file_metadata(
            Path(self.tmp_data_path, self.tmp_dataset_1)).channel_metadata.num_sigs(SignalType.AMP))

    def test_slice_time(self):
        if not self._has_test_data():
            self.skipTest('~/radix/data/radiens_test_data is not set up')


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: datasource.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import common_pb2 as common__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x64\x61tasource.proto\x12\x06\x61llego\x1a\x0c\x63ommon.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"<\n\x11\x44\x61taSourceRequest\x12\x11\n\tdsourceID\x18\x01 \x03(\t\x12\x14\n\x0crenameDsrcID\x18\x02 \x01(\t\"\xc7\x01\n\x18\x44\x61taSourceSetSaveRequest\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x10\n\x08\x62\x61seName\x18\x02 \x01(\t\x12\x11\n\tdsourceID\x18\x03 \x01(\t\x12\x0b\n\x03uID\x18\x04 \x01(\t\x12(\n\x08\x66ileType\x18\x05 \x01(\x0e\x32\x16.allego.RadixFileTypes\x12\x0f\n\x07isForce\x18\x06 \x01(\x08\x12\x17\n\x0fisBackgroundKPI\x18\x07 \x01(\x08\x12\x17\n\x0fparentDsourceID\x18\x08 \x01(\t\"%\n\x11\x44\x61taSourceIDReply\x12\x10\n\x08sortedID\x18\x01 \x03(\t\"\x95\x01\n\x16\x44\x61taSourceSetSaveReply\x12\x11\n\tdsourceID\x18\x01 \x02(\t\x12(\n\x06status\x18\x02 \x01(\x0b\x32\x18.allego.DataSourceStatus\x12\r\n\x05niIDs\x18\x03 \x03(\t\x12/\n\x0f\x61ssociatedDsrcs\x18\x04 \x03(\x0b\x32\x16.allego.FileDescriptor\"\xae\x01\n\x13\x44\x61taSourceStatusMap\x12\x39\n\x07\x64source\x18\x01 \x03(\x0b\x32(.allego.DataSourceStatusMap.DsourceEntry\x12\x12\n\norderedIDs\x18\x02 \x03(\t\x1aH\n\x0c\x44sourceEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\'\n\x05value\x18\x02 \x01(\x0b\x32\x18.allego.DataSourceStatus:\x02\x38\x01\"\xa1\x03\n\x1f\x43opyRemoveDataSourceFileRequest\x12\x45\n\x04many\x18\x01 \x01(\x0b\x32\x35.allego.CopyRemoveDataSourceFileRequest.MultipleFilesH\x00\x12>\n\x03one\x18\x02 \x01(\x0b\x32/.allego.CopyRemoveDataSourceFileRequest.OneFileH\x00\x12\x0f\n\x07isForce\x18\x03 \x01(\x08\x12\x12\n\nisValidate\x18\x04 \x01(\x08\x1at\n\rMultipleFiles\x12#\n\x03src\x18\x01 \x03(\x0b\x32\x16.allego.FileDescriptor\x12\x10\n\x08\x64\x65stPath\x18\x02 \x01(\t\x12,\n\x0c\x64\x65stFileType\x18\x03 \x01(\x0e\x32\x16.allego.RadixFileTypes\x1aT\n\x07OneFile\x12#\n\x03src\x18\x01 \x02(\x0b\x32\x16.allego.FileDescriptor\x12$\n\x04\x64\x65st\x18\x02 \x01(\x0b\x32\x16.allego.FileDescriptorB\x06\n\x04\x64\x65sc\"\xf8\x02\n\x15\x44\x61taSourceFileSetStat\x12\x10\n\x08numFiles\x18\x01 \x02(\x05\x12\x1b\n\x13numBytesPrimaryData\x18\x02 \x02(\x03\x12\x18\n\x10numBytesMetaData\x18\x03 \x02(\x03\x12\x10\n\x08numBytes\x18\x04 \x02(\x03\x12\x16\n\x0eisMetadataFile\x18\x05 \x02(\x08\x12-\n\ttimeStamp\x18\x06 \x02(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x13\n\x0bnumChannels\x18\x07 \x02(\x05\x12\x13\n\x0b\x64urationSec\x18\x08 \x02(\x01\x12\x12\n\ndatasetUID\x18\t \x02(\t\x12\x17\n\x0f\x64\x61tasetChecksum\x18\n \x01(\t\x12\x19\n\x11\x64\x61tasetProvenance\x18\x0b \x03(\t\x12\x12\n\nsampleRate\x18\x0c \x02(\x05\x12\x37\n\x0e\x62\x61\x63kgroundProc\x18\r \x01(\x0b\x32\x1f.allego.BackgroundProcessStatus\"\x82\x02\n\rCpRmMvLsReply\x12\x10\n\x08numFiles\x18\x01 \x02(\x05\x12\x12\n\nnumDsource\x18\x02 \x02(\x05\x12\x10\n\x08numBytes\x18\x03 \x02(\x03\x12\x10\n\x08\x64\x65stPath\x18\x04 \x02(\t\x12\x35\n\x07\x64source\x18\x05 \x03(\x0b\x32$.allego.CpRmMvLsReply.DataSourceInfo\x12\x0b\n\x03msg\x18\x06 \x01(\t\x1a\x63\n\x0e\x44\x61taSourceInfo\x12$\n\x04\x64\x65sc\x18\x01 \x02(\x0b\x32\x16.allego.FileDescriptor\x12+\n\x04stat\x18\x02 \x02(\x0b\x32\x1d.allego.DataSourceFileSetStat\"y\n\x17\x42\x61\x63kgroundProcessStatus\x12\x0b\n\x03uID\x18\x01 \x02(\t\x12\x14\n\x0c\x66racComplete\x18\x02 \x01(\x01\x12\x16\n\x0e\x65lapsedTimeSec\x18\x03 \x01(\x01\x12#\n\x1b\x65stimatedCompletionWallTime\x18\x04 \x01(\t\"\x8b\x01\n\x19MoveDataSourceFileRequest\x12#\n\x03src\x18\x01 \x02(\x0b\x32\x16.allego.FileDescriptor\x12$\n\x04\x64\x65st\x18\x02 \x02(\x0b\x32\x16.allego.FileDescriptor\x12\x0f\n\x07isForce\x18\x03 \x01(\x08\x12\x12\n\nisValidate\x18\x04 \x01(\x08\"\xbf\x01\n\x10\x44\x61taSourceParams\x12\x12\n\nsampleFreq\x18\x01 \x02(\x01\x12\x11\n\ttimeRange\x18\x02 \x03(\x01\x12\x10\n\x08\x64uration\x18\x03 \x02(\x01\x12%\n\x04type\x18\x05 \x02(\x0e\x32\x17.allego.SignalGroupType\x12\x35\n\x0e\x64\x61tasourceType\x18\x06 \x01(\x0e\x32\x1d.allego.RadiensDatasourceType\x12\x14\n\x0c\x64\x61tasourceID\x18\x07 \x01(\t\".\n\x10ImpedanceRequest\x12\x1a\n\x04port\x18\x01 \x02(\x0e\x32\x0c.allego.Port\"\x85\x01\n\tImpedance\x12\x34\n\nimpedances\x18\x01 \x03(\x0b\x32 .allego.Impedance.ImpedanceValue\x1a\x42\n\x0eImpedanceValue\x12\x12\n\nntvChanIdx\x18\x01 \x02(\x05\x12\x0c\n\x04zMag\x18\x02 \x02(\x01\x12\x0e\n\x06zPhase\x18\x03 \x02(\x01*0\n\x15RadiensDatasourceType\x12\x07\n\x03\x41NY\x10\x00\x12\x0e\n\nTIMESERIES\x10\x01\x42\x15Z\x13internal/radix/grpc')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'datasource_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\023internal/radix/grpc'
  _DATASOURCESTATUSMAP_DSOURCEENTRY._options = None
  _DATASOURCESTATUSMAP_DSOURCEENTRY._serialized_options = b'8\001'
  _RADIENSDATASOURCETYPE._serialized_start=2410
  _RADIENSDATASOURCETYPE._serialized_end=2458
  _DATASOURCEREQUEST._serialized_start=75
  _DATASOURCEREQUEST._serialized_end=135
  _DATASOURCESETSAVEREQUEST._serialized_start=138
  _DATASOURCESETSAVEREQUEST._serialized_end=337
  _DATASOURCEIDREPLY._serialized_start=339
  _DATASOURCEIDREPLY._serialized_end=376
  _DATASOURCESETSAVEREPLY._serialized_start=379
  _DATASOURCESETSAVEREPLY._serialized_end=528
  _DATASOURCESTATUSMAP._serialized_start=531
  _DATASOURCESTATUSMAP._serialized_end=705
  _DATASOURCESTATUSMAP_DSOURCEENTRY._serialized_start=633
  _DATASOURCESTATUSMAP_DSOURCEENTRY._serialized_end=705
  _COPYREMOVEDATASOURCEFILEREQUEST._serialized_start=708
  _COPYREMOVEDATASOURCEFILEREQUEST._serialized_end=1125
  _COPYREMOVEDATASOURCEFILEREQUEST_MULTIPLEFILES._serialized_start=915
  _COPYREMOVEDATASOURCEFILEREQUEST_MULTIPLEFILES._serialized_end=1031
  _COPYREMOVEDATASOURCEFILEREQUEST_ONEFILE._serialized_start=1033
  _COPYREMOVEDATASOURCEFILEREQUEST_ONEFILE._serialized_end=1117
  _DATASOURCEFILESETSTAT._serialized_start=1128
  _DATASOURCEFILESETSTAT._serialized_end=1504
  _CPRMMVLSREPLY._serialized_start=1507
  _CPRMMVLSREPLY._serialized_end=1765
  _CPRMMVLSREPLY_DATASOURCEINFO._serialized_start=1666
  _CPRMMVLSREPLY_DATASOURCEINFO._serialized_end=1765
  _BACKGROUNDPROCESSSTATUS._serialized_start=1767
  _BACKGROUNDPROCESSSTATUS._serialized_end=1888
  _MOVEDATASOURCEFILEREQUEST._serialized_start=1891
  _MOVEDATASOURCEFILEREQUEST._serialized_end=2030
  _DATASOURCEPARAMS._serialized_start=2033
  _DATASOURCEPARAMS._serialized_end=2224
  _IMPEDANCEREQUEST._serialized_start=2226
  _IMPEDANCEREQUEST._serialized_end=2272
  _IMPEDANCE._serialized_start=2275
  _IMPEDANCE._serialized_end=2408
  _IMPEDANCE_IMPEDANCEVALUE._serialized_start=2342
  _IMPEDANCE_IMPEDANCEVALUE._serialized_end=2408
# @@protoc_insertion_point(module_scope)
